@@ -55,13 +55,13 @@ DO_COMMAND(do_chat)
 
 	if (*cmd == 0)
 	{
-		tintin_header(NULL, " CHAT COMMANDS ");
+		tintin_header(ses, " CHAT COMMANDS ");
 
 		for (cnt = 0 ; *chat_table[cnt].name != 0 ; cnt++)
 		{
 			tintin_printf2(ses, "  [%-13s] %s", chat_table[cnt].name, chat_table[cnt].desc);
 		}
-		tintin_header(NULL, "");
+		tintin_header(ses, "");
 
 		return ses;
 	}
@@ -81,8 +81,9 @@ DO_COMMAND(do_chat)
 		}
 
 		arg = get_arg_in_braces(arg, left,  chat_table[cnt].lval);
+		substitute(ses, left, left, SUB_VAR|SUB_FUN);
 		arg = get_arg_in_braces(arg, right, chat_table[cnt].rval);
-
+		substitute(ses, right, right, SUB_VAR|SUB_FUN);
 		chat_table[cnt].fun(left, right);
 
 		return ses;
@@ -105,7 +106,7 @@ DO_CHAT(chat_initialize)
 	struct hostent *hp = NULL;
 	struct linger ld;
 	char *reuse = "1";
-	int sock;
+	int sock, port;
 
 	if (gtd->chat)
 	{
@@ -114,15 +115,7 @@ DO_CHAT(chat_initialize)
 		return;
 	}
 
-	gtd->chat = calloc(1, sizeof(struct chat_data));
-
-	gtd->chat->port     = *left ? atoi(left) : DEFAULT_PORT;
-	gtd->chat->name     = strdup("TinTin");
-	gtd->chat->ip       = strdup("<Unknown>");
-	gtd->chat->reply    = strdup("");
-	gtd->chat->color    = strdup("\033[0;1;31m");
-
-	chat_downloaddir("", "");
+	port = atoi(left) ? atoi(left) : DEFAULT_PORT;
 
 	gethostname(hostname, BUFFER_SIZE);
 
@@ -136,7 +129,7 @@ DO_CHAT(chat_initialize)
 	}
 
 	sa.sin_family      = hp->h_addrtype;
-	sa.sin_port        = htons(gtd->chat->port);
+	sa.sin_port        = htons(port);
 	sa.sin_addr.s_addr = 0;
 
 	sock = socket(AF_INET, SOCK_STREAM, 0);
@@ -166,7 +159,7 @@ DO_CHAT(chat_initialize)
 
 	if (bind(sock, (struct sockaddr *) &sa, sizeof(sa)) < 0)
 	{
-		chat_printf("Port in use, cannot initiate chat.");
+		tintin_printf(NULL, "Port %d is already in use, cannot initiate chat.", port);
 
 		close(sock);
 
@@ -181,7 +174,16 @@ DO_CHAT(chat_initialize)
 
 		return;
 	}
-	gtd->chat->fd = sock;
+	gtd->chat = (struct chat_data *) calloc(1, sizeof(struct chat_data));
+
+	gtd->chat->fd       = sock;
+	gtd->chat->port     = port;
+
+	gtd->chat->color    = strdup("\033[0;1;31m");
+	gtd->chat->download = strdup("");
+	gtd->chat->ip       = strdup("<Unknown>");
+	gtd->chat->name     = strdup("TinTin");
+	gtd->chat->reply    = strdup("");
 
 	chat_printf("Initialized chat on port %d.", gtd->chat->port);
 }
@@ -211,7 +213,7 @@ int chat_new(int s)
 {
 	struct chat_data *new_buddy;
 	struct sockaddr_in sock;
-	unsigned int i;
+	socklen_t i;
 	int fd;
 
 	i = sizeof(sock);
@@ -237,12 +239,15 @@ int chat_new(int s)
 		return -1;
 	}
 
-	new_buddy = calloc(1, sizeof(struct chat_data));
+	new_buddy = (struct chat_data *) calloc(1, sizeof(struct chat_data));
 
-	new_buddy->ip    = strdup(inet_ntoa(sock.sin_addr));
-	new_buddy->fd    = fd;
-	new_buddy->name  = strdup("Unknown");
-	new_buddy->group = strdup("");
+	new_buddy->fd       = fd;
+
+	new_buddy->download = strdup("");
+	new_buddy->group    = strdup("");
+	new_buddy->ip       = strdup(inet_ntoa(sock.sin_addr));
+	new_buddy->name     = strdup("Unknown");
+	new_buddy->version  = strdup("");
 
 	new_buddy->timeout = CALL_TIMEOUT + time(NULL);
 
@@ -355,11 +360,14 @@ void *threaded_chat_call(void *arg)
 
 	new_buddy = calloc(1, sizeof(struct chat_data));
 
-	new_buddy->ip    = strdup(host);
-	new_buddy->port  = atoi(port);
-	new_buddy->fd    = sock;
-	new_buddy->name  = strdup("");
-	new_buddy->group = strdup("");
+	new_buddy->fd       = sock;
+	new_buddy->port     = atoi(port);
+
+	new_buddy->group    = strdup("");
+	new_buddy->ip       = strdup(host);
+	new_buddy->name     = strdup("");
+	new_buddy->version  = strdup("");
+	new_buddy->download = strdup("");
 
 	strip_vt102_codes(gtd->chat->name, name);
 
@@ -509,13 +517,16 @@ void *threaded_chat_call(void *arg)
 		return NULL;
 	}
 
-	new_buddy = calloc(1, sizeof(struct chat_data));
+	new_buddy = (struct chat_data *) calloc(1, sizeof(struct chat_data));
 
-	new_buddy->ip    = strdup(host);
-	new_buddy->port  = atoi(port);
-	new_buddy->fd    = sock;
-	new_buddy->name  = strdup("");
-	new_buddy->group = strdup("");
+	new_buddy->fd       = sock;
+	new_buddy->port     = atoi(port);
+
+	new_buddy->download = strdup("");
+	new_buddy->group    = strdup("");
+	new_buddy->ip       = strdup(host);
+	new_buddy->name     = strdup("");
+	new_buddy->version  = strdup("");
 
 	strip_vt102_codes(gtd->chat->name, name);
 
@@ -574,7 +585,7 @@ DO_CHAT(chat_call)
 
 	sprintf(buf, "{%s} {%s}", left, right);
 
-	pthread_create(&thread, NULL, (void *) threaded_chat_call, (void *) buf);
+	pthread_create(&thread, NULL, threaded_chat_call, (void *) buf);
 }
 	
 #else
@@ -615,14 +626,14 @@ void close_chat(struct chat_data *buddy, int unlink)
 		}
 	}
 
-	STRFREE(buddy->name);
-	STRFREE(buddy->ip);
-	STRFREE(buddy->download);
-	STRFREE(buddy->version);
-	STRFREE(buddy->group);
-
 	close(buddy->fd);
 
+	STRFREE(buddy->download);
+	STRFREE(buddy->group);
+	STRFREE(buddy->ip);
+	STRFREE(buddy->name);
+	STRFREE(buddy->version);
+
 	free(buddy);
 }
 
@@ -719,7 +730,7 @@ void chat_printf(char *format, ...)
 
 	for (buddy = gtd->chat->next ; buddy ; buddy = buddy->next)
 	{
-		if (HAS_BIT(buddy->flags, CHAT_FLAG_FORWARD))
+		if (HAS_BIT(buddy->flags, CHAT_FLAG_FORWARD) && !HAS_BIT(buddy->flags, CHAT_FLAG_FORWARDALL))
 		{
 			chat_socket_printf(buddy, "%s", tmp);
 		}
@@ -938,12 +949,9 @@ void get_chat_commands(struct chat_data *buddy, char *buf, int len)
 
 		*pto-- = 0;
 
-		if (strlen(txt))
+		while (isspace(*pto))
 		{
-			while (isspace(*pto))
-			{
-				*pto-- = 0;
-			}
+			*pto-- = 0;
 		}
 
 		switch (ptc)
@@ -993,11 +1001,11 @@ void get_chat_commands(struct chat_data *buddy, char *buf, int len)
 				break;
 
 			case CHAT_VERSION:
-				if (buddy->version == NULL)
+				if (*buddy->version == 0 && *txt != 0)
 				{
 					chat_socket_printf(buddy, "%c%s%s%c", CHAT_VERSION, "TinTin++ ", VERSION_NUM, CHAT_END_OF_COMMAND);
 
-					buddy->version = strdup(txt);
+					RESTRING(buddy->version, txt);
 				}
 				break;
 
@@ -1043,6 +1051,8 @@ void get_chat_commands(struct chat_data *buddy, char *buf, int len)
 
 			case CHAT_SNOOP_DATA:
 				SET_BIT(buddy->flags, CHAT_FLAG_FORWARDBY);
+				DEL_BIT(buddy->flags, CHAT_FLAG_FORWARD);
+				DEL_BIT(buddy->flags, CHAT_FLAG_FORWARDALL);
 				chat_receive_snoop_data(buddy, txt);
 				break;
 
@@ -1313,8 +1323,6 @@ void parse_requested_connections(struct chat_data *buddy, char *txt)
 		return;
 	}
 
-
-
 	ip[0] = 0;
 	port[0] = 0;
 	comma = txt;
@@ -1376,14 +1384,7 @@ DO_CHAT(chat_downloaddir)
 {
 	char dir[BUFFER_SIZE];
 
-	if (gtd->chat->download == NULL)
-	{
-		gtd->chat->download = strdup("");
-
-		return;
-	}
-
-	sprintf(dir, "%s%s", left, is_suffix("/", left) ? "/" : "");
+	sprintf(dir, "%s%s", left, !str_suffix(left, "/") ? "" : "/");
 
 	RESTRING(gtd->chat->download, dir);
 
@@ -2248,7 +2249,7 @@ DO_CHAT(chat_color)
 {
 	if (*left == 0 || get_highlight_codes(gtd->ses, left, right) == FALSE)
 	{
-		chat_printf("Valid colors are:\r\n\r\nreset, bold, dim, light, dark, underscore, blink, reverse, black, red, green, yellow, blue, magenta, cyan, white, b black, b red, b green, b yellow, b blue, b magenta, b cyan, b white");
+		chat_printf("Valid colors are:\n\nreset, bold, dim, light, dark, underscore, blink, reverse, black, red, green, yellow, blue, magenta, cyan, white, b black, b red, b green, b yellow, b blue, b magenta, b cyan, b white");
 
 		return;
 	}
@@ -2427,7 +2428,7 @@ void chat_forward_session(struct session *ses, char *linelog)
 }
 
 /*
-	Pussy version of the ignore feature
+	The ignore feature.
 */
 
 DO_CHAT(chat_ignore)
