@@ -34,38 +34,41 @@ DO_COMMAND(do_configure)
 	char left[BUFFER_SIZE], right[BUFFER_SIZE];
 	struct listroot *root;
 	struct listnode *node;
-	int cnt;
+	int index;
 
 	root = ses->list[LIST_CONFIG];
 
 	arg = get_arg_in_braces(arg, left,  FALSE);
-	arg = get_arg_in_braces(arg, right, FALSE);
+	arg = sub_arg_in_braces(ses, arg, right, GET_ONE, SUB_VAR|SUB_FUN);
 
 	if (*left == 0)
 	{
 		tintin_header(ses, " CONFIGURATIONS ");
 
-		node = root->f_node;
-
-		for (node = root->f_node ; node ; node = node->next)
+		for (index = 0 ; *config_table[index].name != 0 ; index++)
 		{
-			tintin_printf2(ses, "[%-13s] [%8s] %s", 
-				node->left,
-				node->right,
-				strcmp(node->right, "ON") == 0 ? config_table[atoi(node->pr)].msg_on : config_table[atoi(node->pr)].msg_off);
+			node = search_node_list(ses->list[LIST_CONFIG], config_table[index].name);
+
+			if (node)
+			{
+				tintin_printf2(ses, "[%-13s] [%8s] %s", 
+					node->left,
+					node->right,
+					strcmp(node->right, "ON") == 0 ? config_table[index].msg_on : config_table[index].msg_off);
+			}
 		}
 
 		tintin_header(ses, "");
 	}
 	else
 	{
-		for (cnt = 0 ; *config_table[cnt].name != 0 ; cnt++)
+		for (index = 0 ; *config_table[index].name != 0 ; index++)
 		{
-			if (is_abbrev(left, config_table[cnt].name))
+			if (is_abbrev(left, config_table[index].name))
 			{
-				if (config_table[cnt].config(ses, right, cnt) != NULL)
+				if (config_table[index].config(ses, right, index) != NULL)
 				{
-					show_message(ses, LIST_CONFIG, "#CONFIG {%s} HAS BEEN SET TO {%s}.", config_table[cnt].name, capitalize(right));
+					show_message(ses, LIST_CONFIG, "#CONFIG {%s} HAS BEEN SET TO {%s}.", config_table[index].name, capitalize(right));
 				}
 				return ses;
 			}
@@ -78,8 +81,6 @@ DO_COMMAND(do_configure)
 
 DO_CONFIG(config_speedwalk)
 {
-	char str[BUFFER_SIZE];
-
 	if (!strcasecmp(arg, "ON"))
 	{
 		SET_BIT(ses->flags, SES_FLAG_SPEEDWALK);
@@ -94,9 +95,7 @@ DO_CONFIG(config_speedwalk)
 
 		return NULL;
 	}
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
@@ -104,8 +103,6 @@ DO_CONFIG(config_speedwalk)
 
 DO_CONFIG(config_verbatim)
 {
-	char str[BUFFER_SIZE];
-
 	if (!strcasecmp(arg, "ON"))
 	{
 		SET_BIT(ses->flags, SES_FLAG_VERBATIM);
@@ -120,17 +117,13 @@ DO_CONFIG(config_verbatim)
 
 		return NULL;
 	}
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
 
 DO_CONFIG(config_repeatenter)
 {
-	char str[BUFFER_SIZE];
-
 	if (!strcasecmp(arg, "ON"))
 	{
 		SET_BIT(ses->flags, SES_FLAG_REPEATENTER);
@@ -145,16 +138,14 @@ DO_CONFIG(config_repeatenter)
 
 		return NULL;
 	}
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
 
 DO_CONFIG(config_commandcolor)
 {
-	char str[BUFFER_SIZE];
+	char buf[BUFFER_SIZE];
 
 	if (!is_color_code(arg))
 	{
@@ -162,19 +153,17 @@ DO_CONFIG(config_commandcolor)
 
 		return NULL;
 	}
-	substitute(ses, arg, ses->command_color, SUB_COL);
+	substitute(ses, arg, buf, SUB_COL);
 
-	sprintf(str, "%d", index);
+	RESTRING(ses->cmd_color, buf);
 
-	updatenode_list(ses, config_table[index].name, arg, str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, arg, "");
 
 	return ses;
 }
 
 DO_CONFIG(config_commandecho)
 {
-	char str[BUFFER_SIZE];
-
 	if (!strcasecmp(arg, "ON"))
 	{
 		SET_BIT(ses->flags, SES_FLAG_ECHOCOMMAND);
@@ -189,17 +178,13 @@ DO_CONFIG(config_commandecho)
 
 		return NULL;
 	}
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
 
 DO_CONFIG(config_verbose)
 {
-	char str[BUFFER_SIZE];
-
 	if (!strcasecmp(arg, "ON"))
 	{
 		SET_BIT(ses->flags, SES_FLAG_VERBOSE);
@@ -214,17 +199,13 @@ DO_CONFIG(config_verbose)
 
 		return NULL;
 	}
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
 
 DO_CONFIG(config_wordwrap)
 {
-	char str[BUFFER_SIZE];
-
 	if (!strcasecmp(arg, "ON"))
 	{
 		SET_BIT(ses->flags, SES_FLAG_WORDWRAP);
@@ -239,17 +220,13 @@ DO_CONFIG(config_wordwrap)
 
 		return NULL;
 	}
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
 
 DO_CONFIG(config_log)
 {
-	char str[BUFFER_SIZE];
-
 	if (!strcasecmp(arg, "HTML"))
 	{
 		DEL_BIT(ses->flags, SES_FLAG_LOGPLAIN);
@@ -271,17 +248,13 @@ DO_CONFIG(config_log)
 
 		return NULL;
 	}
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
 
 DO_CONFIG(config_buffersize)
 {
-	char str[BUFFER_SIZE];
-
 	if (!is_number(arg))
 	{
 		tintin_printf(ses, "#SYNTAX: #CONFIG {BUFFER SIZE} <NUMBER>");
@@ -298,17 +271,13 @@ DO_CONFIG(config_buffersize)
 
 	init_buffer(ses, atoi(arg));
 
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
 
 DO_CONFIG(config_scrolllock)
 {
-	char str[BUFFER_SIZE];
-
 	if (!strcasecmp(arg, "ON"))
 	{
 		SET_BIT(ses->flags, SES_FLAG_SCROLLLOCK);
@@ -323,17 +292,13 @@ DO_CONFIG(config_scrolllock)
 
 		return NULL;
 	}
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
 
 DO_CONFIG(config_connectretry)
 {
-	char str[BUFFER_SIZE];
-
 	if (!is_number(arg))
 	{
 		tintin_printf(ses, "#SYNTAX: #CONFIG {CONNECT RETRY} <NUMBER>");
@@ -343,17 +308,13 @@ DO_CONFIG(config_connectretry)
 
 	gts->connect_retry = atoll(arg) * 1000000LL;
 
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
 
 DO_CONFIG(config_packetpatch)
 {
-	char str[BUFFER_SIZE];
-
 	if (!is_number(arg))
 	{
 		tintin_printf(ses, "#SYNTAX: #CONFIG {PACKET PATCH} <NUMBER>");
@@ -361,7 +322,7 @@ DO_CONFIG(config_packetpatch)
 		return NULL;
 	}
 
-	if (atof(arg) < 0 || atoi(arg) > 10)
+	if (atof(arg) < 0 || atof(arg) > 10)
 	{
 		tintin_printf(ses, "#ERROR: #CONFIG PACKET PATCH: PROVIDE A NUMBER BETWEEN 0.00 and 10.00");
 
@@ -370,9 +331,7 @@ DO_CONFIG(config_packetpatch)
 
 	gts->check_output = (long long) (get_number(ses, arg) * 1000000LL);
 
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
@@ -380,8 +339,6 @@ DO_CONFIG(config_packetpatch)
 
 DO_CONFIG(config_historysize)
 {
-	char str[BUFFER_SIZE];
-
 	if (!is_number(arg))
 	{
 		tintin_printf(ses, "#SYNTAX: #CONFIG {HISTORY SIZE} <NUMBER>");
@@ -396,56 +353,40 @@ DO_CONFIG(config_historysize)
 
 	gtd->history_size = atoi(arg);
 
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
 
 DO_CONFIG(config_tintinchar)
 {
-	char str[BUFFER_SIZE];
-
 	gtd->tintin_char = arg[0];
 
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, arg, str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, arg, "");
 
 	return ses;
 }
 
 DO_CONFIG(config_verbatimchar)
 {
-	char str[BUFFER_SIZE];
-
 	gtd->verbatim_char = arg[0];
 
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, arg, str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, arg, "");
 
 	return ses;
 }
 
 DO_CONFIG(config_repeatchar)
 {
-	char str[BUFFER_SIZE];
-
 	gtd->repeat_char = arg[0];
 
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, arg, str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, arg, "");
 
 	return ses;
 }
 
 DO_CONFIG(config_debugtelnet)
 {
-	char str[BUFFER_SIZE];
-
 	if (!strcasecmp(arg, "ON"))
 	{
 		SET_BIT(ses->telopts, TELOPT_FLAG_DEBUG);
@@ -460,17 +401,13 @@ DO_CONFIG(config_debugtelnet)
 
 		return NULL;
 	}
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
 
 DO_CONFIG(config_convertmeta)
 {
-	char str[BUFFER_SIZE];
-
 	if (!strcasecmp(arg, "ON"))
 	{
 		SET_BIT(ses->flags, SES_FLAG_CONVERTMETA);
@@ -485,17 +422,13 @@ DO_CONFIG(config_convertmeta)
 
 		return NULL;
 	}
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
 
 DO_CONFIG(config_loglevel)
 {
-	char str[BUFFER_SIZE];
-
 	if (!strcasecmp(arg, "LOW"))
 	{
 		SET_BIT(ses->flags, SES_FLAG_LOGLEVEL);
@@ -510,17 +443,13 @@ DO_CONFIG(config_loglevel)
 
 		return NULL;
 	}
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
 
 DO_CONFIG(config_colorpatch)
 {
-	char str[BUFFER_SIZE];
-
 	if (!strcasecmp(arg, "ON"))
 	{
 		SET_BIT(ses->flags, SES_FLAG_COLORPATCH);
@@ -535,24 +464,20 @@ DO_CONFIG(config_colorpatch)
 
 		return NULL;
 	}
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
 
-DO_CONFIG(config_regexp)
+DO_CONFIG(config_mccp)
 {
-	char str[BUFFER_SIZE];
-
 	if (!strcasecmp(arg, "ON"))
 	{
-		SET_BIT(ses->flags, SES_FLAG_REGEXP);
+		SET_BIT(ses->flags, SES_FLAG_MCCP);
 	}
 	else if (!strcasecmp(arg, "OFF"))
 	{
-		DEL_BIT(ses->flags, SES_FLAG_REGEXP);
+		DEL_BIT(ses->flags, SES_FLAG_MCCP);
 	}
 	else
 	{
@@ -560,24 +485,31 @@ DO_CONFIG(config_regexp)
 
 		return NULL;
 	}
-	sprintf(str, "%d", index);
-
-	updatenode_list(ses, config_table[index].name, capitalize(arg), str, LIST_CONFIG);
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
 
-DO_CONFIG(config_timestamp)
+DO_CONFIG(config_autotab)
 {
-	char str[BUFFER_SIZE];
+	if (!is_number(arg))
+	{
+		tintin_printf(ses, "#SYNTAX: #CONFIG {AUTO TAB} <NUMBER>");
 
-	substitute(ses, arg, str, SUB_COL|SUB_ESC);
+		return NULL;
+	}
 
-	RESTRING(ses->timestamp, str);
+	if (atoi(arg) < 1 || atoi(arg) > 999999)
+	{
+		tintin_printf(ses, "#ERROR: #CONFIG BUFFER: PROVIDE A NUMBER BETWEEN 1 and 999999");
 
-	sprintf(str, "%d", index);
+		return NULL;
+	}
 
-	updatenode_list(ses, config_table[index].name, *ses->timestamp ? "ON" : "OFF", str, LIST_CONFIG);
+	ses->auto_tab = atoi(arg);
+
+	update_node_list(ses->list[LIST_CONFIG], config_table[index].name, capitalize(arg), "");
 
 	return ses;
 }
+
