@@ -29,8 +29,9 @@
 
 void process_input(void)
 {
-	if (HAS_BIT(gtd->ses->telopts, TELOPT_FLAG_SGA)
-	&& !HAS_BIT(gtd->ses->telopts, TELOPT_FLAG_ECHO))
+	int flags;
+
+	if (HAS_BIT(gtd->ses->telopts, TELOPT_FLAG_SGA)	&& !HAS_BIT(gtd->ses->telopts, TELOPT_FLAG_ECHO))
 	{
 		read_key();
 	}
@@ -55,7 +56,7 @@ void process_input(void)
 
 	if (HAS_BIT(gtd->ses->telopts, TELOPT_FLAG_ECHO))
 	{
-		add_line_history(gts, gtd->input_buf);
+		add_line_history(gtd->ses, gtd->input_buf);
 	}
 
 	if (HAS_BIT(gtd->ses->telopts, TELOPT_FLAG_ECHO))
@@ -69,16 +70,18 @@ void process_input(void)
 
 	if (gtd->ses->scroll_line != -1)
 	{
-		buffer_e("");
+		buffer_end(gtd->ses, "");
 	}
 
-	SET_BIT(gtd->flags, TINTIN_FLAG_USERCOMMAND|TINTIN_FLAG_SHOWMESSAGE);
+	flags = TINTIN_FLAG_USERCOMMAND | (HAS_BIT(gtd->ses->flags, SES_FLAG_VERBATIM) ? TINTIN_FLAG_VERBATIM : 0);
 
-	pre_parse_input(gtd->ses, gtd->input_buf, SUB_NONE);
+	SET_BIT(gtd->flags, flags);
 
-	DEL_BIT(gtd->flags, TINTIN_FLAG_USERCOMMAND|TINTIN_FLAG_SHOWMESSAGE);
+	gtd->ses = script_driver(gtd->ses, -1, gtd->input_buf);
 
-	check_all_events(gtd->ses, "RECEIVED INPUT");
+	DEL_BIT(gtd->flags, flags);
+
+	check_all_events(gtd->ses, 0, 1, "RECEIVED INPUT", gtd->input_buf);
 
 	if (IS_SPLIT(gtd->ses))
 	{
@@ -115,13 +118,13 @@ void read_line()
 		match = 0;
 		root  = gtd->ses->list[LIST_MACRO];
 
-		for (node = root->f_node ; node ; node = root->update)
+		for (root->update = 0 ; root->update < root->used ; root->update++)
 		{
-			root->update = node->next;
+			node = root->list[root->update];
 
 			if (!strcmp(gtd->macro_buf, node->pr))
 			{
-				pre_parse_input(gtd->ses, node->right, SUB_NONE);
+				script_driver(gtd->ses, LIST_MACRO, node->right);
 
 				gtd->macro_buf[0] = 0;
 
@@ -203,6 +206,10 @@ void read_line()
 
 				cursor_check_line("");
 
+				DEL_BIT(gtd->flags, TINTIN_FLAG_HISTORYBROWSE);
+
+				kill_list(gtd->ses->list[LIST_TABCYCLE]);
+
 				if (HAS_BIT(gtd->flags, TINTIN_FLAG_HISTORYSEARCH))
 				{
 					cursor_history_find("");
@@ -245,13 +252,13 @@ void read_key(void)
 	
 		root  = gtd->ses->list[LIST_MACRO];
 
-		for (node = root->f_node ; node ; node = root->update)
+		for (root->update = 0 ; root->update < root->used ; root->update++)
 		{
-			root->update = node->next;
+			node = root->list[root->update];
 
 			if (!strcmp(gtd->macro_buf, node->pr))
 			{
-				pre_parse_input(gtd->ses, node->right, SUB_NONE);
+				script_driver(gtd->ses, LIST_MACRO, node->right);
 
 				gtd->macro_buf[0] = 0;
 				return;
@@ -273,10 +280,22 @@ void read_key(void)
 		switch (gtd->macro_buf[cnt])
 		{
 			case '\r':
+				printf("\n\[1;31mfound \\r!\n");
+				break;
+
 			case '\n':
 				gtd->input_buf[0] = 0;
 				gtd->macro_buf[0] = 0;
-				socket_printf(gtd->ses, 1, "%c", '\r'); 
+				gtd->input_len = 0;
+
+				if (HAS_BIT(gtd->ses->flags, SES_FLAG_RUN))
+				{
+					socket_printf(gtd->ses, 1, "%c", '\r');
+				}
+				else
+				{
+					socket_printf(gtd->ses, 2, "%c%c", '\r', '\n');
+				}
 				break;
 
 			default:
@@ -334,18 +353,47 @@ void convert_meta(char *input, char *output)
 				pti++;
 				break;
 
-			case '\n':
+			case '\a':
+				*pto++ = '\\';
+				*pto++ = 'a';
+				pti++;
+				break;
+
+			case '\b':
+				*pto++ = '\\';
+				*pto++ = 'b';
+				pti++;
+				break;
+
+			case '\t':
+				*pto++ = '\\';
+				*pto++ = 't';
+				pti++;
+				break;
+
 			case '\r':
+				*pto++ = '\\';
+				*pto++ = 'r';
+				pti++;
+				break;
+
+			case '\n':
 				*pto++ = *pti++;
 				break;
 
 			default:
-				if (*pti > 0 && *pti < 30)
+				if (*pti < 32)
 				{
 					*pto++ = '\\';
-					*pto++ = 'C';
-					*pto++ = '-';
-					*pto++ = 'a' + *pti - 1;
+					*pto++ = 'c';
+					if (*pti <= 26)
+					{
+						*pto++ = 'a' + *pti - 1;
+					}
+					else
+					{
+						*pto++ = 'A' + *pti - 1;
+					}
 					pti++;
 					break;
 				}
@@ -385,6 +433,16 @@ void unconvert_meta(char *input, char *output)
 						}
 						break;
 
+					case 'c':
+						*pto++ = pti[2] % 32;
+						pti += 3;
+						break;
+
+					case 'a':
+						*pto++  = '\a';
+						pti += 2;
+						break;
+
 					case 'b':
 						*pto++  = 127;
 						pti    += 2;
@@ -435,7 +493,7 @@ void echo_command(struct session *ses, char *line, int newline)
 
 	if (HAS_BIT(ses->flags, SES_FLAG_SPLIT))
 	{
-		sprintf(buffer, "%s%s", ses->command_color, line);
+		sprintf(buffer, "%s%s\033[0m", ses->cmd_color, line);
 	}
 	else
 	{
@@ -515,3 +573,19 @@ void input_printf(char *format, ...)
 
 	printf("%s", buf);
 }
+
+void modified_input(void)
+{
+	kill_list(gtd->ses->list[LIST_TABCYCLE]);
+
+	if (HAS_BIT(gtd->flags, TINTIN_FLAG_HISTORYSEARCH))
+	{
+		cursor_history_find("");
+	}
+
+	if (HAS_BIT(gtd->flags, TINTIN_FLAG_HISTORYBROWSE))
+	{
+		DEL_BIT(gtd->flags, TINTIN_FLAG_HISTORYBROWSE);
+	}
+
+}
