@@ -30,28 +30,19 @@
 
 void logit(struct session *ses, char *txt, FILE *file, int newline)
 {
-	char buf[BUFFER_SIZE], out[BUFFER_SIZE];
+	char out[BUFFER_SIZE];
 
-	if (*ses->timestamp)
-	{
-		sprintf(buf, "%s%s", timestamp(ses->timestamp), txt);
-	}
-	else
-	{
-		strcpy(buf, txt);
-	}
-	
 	if (HAS_BIT(ses->flags, SES_FLAG_LOGPLAIN))
 	{
-		strip_vt102_codes(buf, out);
+		strip_vt102_codes(txt, out);
 	}
 	else if (HAS_BIT(ses->flags, SES_FLAG_LOGHTML))
 	{
-		vt102_to_html(ses, buf, out);
+		vt102_to_html(ses, txt, out);
 	}
 	else
 	{
-		strcpy(out, buf);
+		strcpy(out, txt);
 	}
 
 	if (newline)
@@ -69,13 +60,16 @@ DO_COMMAND(do_log)
 	char left[BUFFER_SIZE], right[BUFFER_SIZE];
 
 	arg = get_arg_in_braces(arg, left,  FALSE);
+	substitute(ses, left, left, SUB_VAR|SUB_FUN);
+
 	arg = get_arg_in_braces(arg, right, TRUE);
+	substitute(ses, right, right, SUB_VAR|SUB_FUN);
 
 	if (ses->logfile)
 	{
 		fclose(ses->logfile);
 		ses->logfile = NULL;
-		tintin_printf(ses, "#OK: LOGGING TURNED OFF.");
+		show_message(ses, -1, "#OK: LOGGING TURNED OFF.");
 	}
 	else if (*left == 0 || *right == 0 || (!is_abbrev(left, "APPEND") && !is_abbrev(left, "OVERWRITE")))
 	{
@@ -93,11 +87,11 @@ DO_COMMAND(do_log)
 				{
 					write_html_header(ses->logfile);
 				}
-				tintin_printf(ses, "#OK: LOGGING OUTPUT TO '%s' FILESIZE: %ld", right, ftell(ses->logfile));
+				show_message(ses, -1, "#OK: LOGGING OUTPUT TO '%s' FILESIZE: %ld", right, ftell(ses->logfile));
 			}
 			else
 			{
-				tintin_printf(ses, "#ERROR: #LOG {%s} {%s} - COULDN'T OPEN FILE.", left, right);
+				tintin_printf2(ses, "#ERROR: #LOG {%s} {%s} - COULDN'T OPEN FILE.", left, right);
 			}
 		}
 		else
@@ -108,126 +102,17 @@ DO_COMMAND(do_log)
 				{
 					write_html_header(ses->logfile);
 				}
-				tintin_printf(ses, "#OK: LOGGING OUTPUT TO '%s'", right);
+				show_message(ses, -1, "#OK: LOGGING OUTPUT TO '%s'", right);
 			}
 			else
 			{
-				tintin_printf(ses, "#ERROR: #LOG {%s} {%s} - COULDN'T OPEN FILE.", left, right);
+				tintin_printf2(ses, "#ERROR: #LOG {%s} {%s} - COULDN'T OPEN FILE.", left, right);
 			}
 		}
 	}
 	return ses;
 }
 
-
-DO_COMMAND(do_logline)
-{
-	char left[BUFFER_SIZE], right[BUFFER_SIZE];
-
-	arg = get_arg_in_braces(arg, left,  0);
-	arg = get_arg_in_braces(arg, right, 1);
-
-	substitute(ses, right, right, SUB_ESC|SUB_COL);
-
-	if (ses->logline)
-	{
-		return ses;
-	}
-
-	if ((ses->logline = fopen(left, "a")))
-	{
-		fseek(ses->logline, 0, SEEK_END);
-
-		if (ftell(ses->logline) == 0 && HAS_BIT(ses->flags, SES_FLAG_LOGHTML))
-		{
-			write_html_header(ses->logline);
-		}
-
-		if (*right)
-		{
-			logit(ses, right, ses->logline, TRUE);
-
-			fclose(ses->logline);
-			ses->logline = NULL;
-		}
-	}
-	else
-	{
-		tintin_printf(ses, "#ERROR: #LOGLINE {%s} - COULDN'T OPEN FILE.", left);
-	}
-	return ses;
-}
-
-
-DO_COMMAND(do_writebuffer)
-{
-	FILE *fp;
-	char left[BUFFER_SIZE], out[STRING_SIZE];
-	int cnt;
-
-	arg = get_arg_in_braces(arg, left, FALSE);
-
-	if (*left == 0)
-	{
-		tintin_printf(ses, "#SYNTAX: #BUFFER WRITE <FILENAME>]");
-	}
-	else
-	{
-		if ((fp = fopen(left, "w")))
-		{
-			show_message(ses, LIST_MESSAGE, "#OK: WRITING BUFFER TO '%s'.", left);
-
-			if (HAS_BIT(ses->flags, SES_FLAG_LOGHTML))
-			{
-				write_html_header(fp);
-			}
-
-			cnt = ses->scroll_row;
-
-			do
-			{
-				if (cnt == 0)
-				{
-					cnt = ses->scroll_max - 1;
-				}
-				else
-				{
-					cnt--;
-				}
-
-				if (ses->buffer[cnt] == NULL)
-				{
-					continue;
-				}
-
-				if (HAS_BIT(ses->flags, SES_FLAG_LOGPLAIN))
-				{
-					strip_vt102_codes(ses->buffer[cnt], out);
-				}
-				else if (HAS_BIT(ses->flags, SES_FLAG_LOGHTML))
-				{
-					vt102_to_html(ses, ses->buffer[cnt], out);
-				}
-				else
-				{
-					strcpy(out, ses->buffer[cnt]);
-				}
-				strcat(out, "\n");
-
-				fputs(out, fp);
-			}
-			while (cnt != ses->scroll_row);
-
-			fclose(fp);
-		}
-		else
-		{
-			tintin_printf(ses, "#ERROR: #BUFFER WRITE {%s} - FAILED TO OPEN FILE.", left);
-		}
-	}
-	return ses;
-}
-
 void write_html_header(FILE *fp)
 {
 	char *header =
@@ -237,16 +122,16 @@ void write_html_header(FILE *fp)
 		"<meta name='description' content='Generated by TinTin++ "VERSION_NUM" - http://tintin.sourceforge.net'>\n"
 		"<style type='text/css'>\n"
 		"{\n\tfont-family: Courier;\n\tfont-size: 10pt;\n}\n"
-		".d30{ color: #000; } .l30{ color: #555; } .b40{ background-color: #000; }\n"
-		".d31{ color: #B00; } .l31{ color: #F55; } .b41{ background-color: #B00; }\n"
-		".d32{ color: #0B0; } .l32{ color: #5F5; } .b42{ background-color: #0B0; }\n"
-		".d33{ color: #BB0; } .l33{ color: #FF5; } .b43{ background-color: #BB0; }\n"
-		".d34{ color: #00B; } .l34{ color: #55F; } .b44{ background-color: #00B; }\n"
-		".d35{ color: #B0B; } .l35{ color: #F5F; } .b45{ background-color: #B0B; }\n"
-		".d36{ color: #0BB; } .l36{ color: #5FF; } .b46{ background-color: #0BB; }\n"
-		".d37{ color: #BBB; } .l37{ color: #FFF; } .b47{ background-color: #BBB; }\n"
-		".d38{ color: #FFF; } .l38{ color: #FFF; } .b48{ background-color: #000; }\n"
-		".d39{ color: #FFF; } .l39{ color: #FFF; } .b49{ background-color: #000; }\n"
+		".d30{ color: #000; } .l30{ color: #555; } .b40{ background-color: #000; } .b50{ background-color: #555 }\n"
+		".d31{ color: #B00; } .l31{ color: #F55; } .b41{ background-color: #B00; } .b51{ background-color: #F55 }\n"
+		".d32{ color: #0B0; } .l32{ color: #5F5; } .b42{ background-color: #0B0; } .b52{ background-color: #5F5 }\n"
+		".d33{ color: #BB0; } .l33{ color: #FF5; } .b43{ background-color: #BB0; } .b53{ background-color: #FF5 }\n"
+		".d34{ color: #00B; } .l34{ color: #55F; } .b44{ background-color: #00B; } .b54{ background-color: #55F }\n"
+		".d35{ color: #B0B; } .l35{ color: #F5F; } .b45{ background-color: #B0B; } .b55{ background-color: #F5F }\n"
+		".d36{ color: #0BB; } .l36{ color: #5FF; } .b46{ background-color: #0BB; } .b56{ background-color: #5FF }\n"
+		".d37{ color: #BBB; } .l37{ color: #FFF; } .b47{ background-color: #BBB; } .b57{ background-color: #FFF }\n"
+		".d38{ color: #FFF; } .l38{ color: #FFF; } .b48{ background-color: #000; } .b58{ background-color: #000 }\n"
+		".d39{ color: #FFF; } .l39{ color: #FFF; } .b49{ background-color: #000; } .b59{ background-color: #000 }\n"
 		"</style>\n"
 		"<body bgcolor='#000000'>\n"
 		"</head>\n"
@@ -258,7 +143,7 @@ void write_html_header(FILE *fp)
 
 void vt102_to_html(struct session *ses, char *txt, char *out)
 {
-	char tmp[BUFFER_SIZE], *pti, *pto;
+	char tmp[BUFFER_SIZE], *pti, *pto, x[] = { '0', '5', '8', 'B', 'D', 'F' };
 	int vtc, fgc, bgc, cnt;
 
 	vtc = ses->vtc;
@@ -291,70 +176,88 @@ void vt102_to_html(struct session *ses, char *txt, char *out)
 						cnt = -1;
 						pti += 1 + strlen(tmp);
 
-						switch (atoi(tmp))
+						if (HAS_BIT(vtc, COL_256) && (HAS_BIT(vtc, COL_XTF) || HAS_BIT(vtc, COL_XTB)))
 						{
-							case 0:
-								vtc = 0;
-								fgc = 39;
-								bgc = 49;
-								break;
-							case 1:
-								SET_BIT(vtc, COL_BLD);
-								break;
-							case 4:
-								SET_BIT(vtc, COL_UND);
-								break;
-							case 5:
-								break;
-							case 7:
-								SET_BIT(vtc, COL_REV);
-								break;
-							case  2:
-							case 21:
-							case 22:
-								DEL_BIT(vtc, COL_BLD);
-								break;
-							case 24:
-								DEL_BIT(vtc, COL_UND);
-								break;
-							case 27:
-								DEL_BIT(vtc, COL_REV);
-								break;
-							case 38:
-								SET_BIT(vtc, COL_UND);
-								fgc = 38;
-								break;
-							case 39:
-								DEL_BIT(vtc, COL_UND);
-								fgc = 39;
-								break;
-							default:
-								if (atoi(tmp) >= 40 && atoi(tmp) < 50)
-								{
-									bgc = atoi(tmp);
-								}
-								if (atoi(tmp) >= 30 && atoi(tmp) < 40)
-								{
-									fgc = atoi(tmp);
-								}
-								break;
+							if (HAS_BIT(vtc, COL_XTF))
+							{
+								fgc = URANGE(0, atoi(tmp), 255);
+							}
+
+							if (HAS_BIT(vtc, COL_XTB))
+							{
+								bgc = URANGE(0, atoi(tmp), 255);
+							}
+						}
+						else
+						{
+							switch (atoi(tmp))
+							{
+								case 0:
+									vtc = 0;
+									fgc = 39;
+									bgc = 49;
+									break;
+								case 1:
+									SET_BIT(vtc, COL_BLD);
+									break;
+								case 5:
+									if (HAS_BIT(vtc, COL_XTF) || HAS_BIT(vtc, COL_XTB))
+									{
+										SET_BIT(vtc, COL_256);
+									}
+									break;
+								case 7:
+									SET_BIT(vtc, COL_REV);
+									break;
+								case  2:
+								case 21:
+								case 22:
+									DEL_BIT(vtc, COL_BLD);
+									break;
+								case 27:
+									DEL_BIT(vtc, COL_REV);
+									break;
+								case 38:
+									DEL_BIT(vtc, COL_XTB);
+									SET_BIT(vtc, COL_XTF);
+									fgc = 38;
+									break;
+								case 48:
+									DEL_BIT(vtc, COL_XTF);
+									SET_BIT(vtc, COL_XTB);
+									bgc = 48;
+									break;
+								default:
+									if (atoi(tmp) / 10 == 4)
+									{
+										bgc = atoi(tmp);
+									}
+									if (atoi(tmp) / 10 == 10)
+									{
+										bgc = atoi(tmp) - 50;
+									}
+
+									if (atoi(tmp) / 10 == 3)
+									{
+										fgc = atoi(tmp);
+									}
+									if (atoi(tmp) / 10 == 9)
+									{
+										SET_BIT(vtc, COL_BLD);
+
+										fgc = atoi(tmp) - 60;
+									}
+									break;
+							}
 						}
 					}
+
 					if (pti[-1] == 'm')
 					{
 						break;
 					}
 				}
-				if (HAS_BIT(vtc, COL_UND) && !HAS_BIT(ses->vtc, COL_UND))
-				{
-					strcpy(pto, "<u>");
-					pto += strlen(pto);
-				}
-				if (!HAS_BIT(vtc, COL_UND) && HAS_BIT(ses->vtc, COL_UND))
-				{
-					strcpy(pto, "</u>");
-					pto += strlen(pto);
-				}
+
 
 				if (!HAS_BIT(vtc, COL_REV) && HAS_BIT(ses->vtc, COL_REV))
 				{
@@ -365,26 +268,68 @@ void vt102_to_html(struct session *ses, char *txt, char *out)
 
 				if (bgc != ses->bgc || fgc != ses->fgc || vtc != ses->vtc)
 				{
-
 					sprintf(pto, "</span>");
 					pto += strlen(pto);
 
 					if (bgc != ses->bgc)
 					{
-						sprintf(pto, "</span><span class='b%d'>", bgc);
+						if (HAS_BIT(vtc, COL_256) && HAS_BIT(vtc, COL_XTB))
+						{
+							if (bgc < 8)
+							{
+								sprintf(pto, "</span><span class='b%d'>", 40+bgc);
+							}
+							else if (bgc < 16)
+							{
+								sprintf(pto, "</span><span class='b%d'>", 50+bgc-8);
+							}
+							else if (bgc < 232)
+							{
+								sprintf(pto, "</span><span style='background-color: #%c%c%c;'>", x[(bgc-16) / 36], x[(bgc-16) % 36 / 6], x[(bgc-16) % 6]);
+							}
+							else
+							{
+								sprintf(pto, "</span><span style='background-color: rgb(%d,%d,%d);'>", (bgc-232) * 10 + 8, (bgc-232) * 10 + 8,(bgc-232) * 10 + 8);
+							}
+						}
+						else
+						{
+							sprintf(pto, "</span><span class='b%d'>", bgc);
+						}
 						pto += strlen(pto);
 					}
 
-					if (HAS_BIT(vtc, COL_BLD))
+					if (HAS_BIT(vtc, COL_256) && HAS_BIT(vtc, COL_XTF))
 					{
-						sprintf(pto, "<span class='l%d'>", fgc);
-						pto += strlen(pto);
+						if (fgc < 8)
+						{
+							sprintf(pto, "</span><span class='d%d'>", 30+fgc);
+						}
+						else if (fgc < 16)
+						{
+							sprintf(pto, "</span><span class='l%d'>", 30+fgc-8);
+						}
+						else if (fgc < 232)
+						{
+							sprintf(pto, "<span style='color: #%c%c%c;'>", x[(fgc-16) / 36], x[(fgc-16) % 36 / 6], x[(fgc-16) % 6]);
+						}
+						else
+						{
+							sprintf(pto, "<span style='color: rgb(%d,%d,%d);'>", (fgc-232) * 10 + 8, (fgc-232) * 10 + 8,(fgc-232) * 10 + 8);
+						}
 					}
 					else
 					{
-						sprintf(pto, "<span class='d%d'>", fgc);
-						pto += strlen(pto);
+						if (HAS_BIT(vtc, COL_BLD))
+						{
+							sprintf(pto, "<span class='l%d'>", fgc);
+						}
+						else
+						{
+							sprintf(pto, "<span class='d%d'>", fgc);
+						}
 					}
+					pto += strlen(pto);
 				}
 
 				if (HAS_BIT(vtc, COL_REV) && !HAS_BIT(ses->vtc, COL_REV))
@@ -393,12 +338,30 @@ void vt102_to_html(struct session *ses, char *txt, char *out)
 					fgc = ses->fgc = bgc - 10;
 					bgc = ses->bgc = cnt + 10;
 				}
+				DEL_BIT(vtc, COL_XTF);
+				DEL_BIT(vtc, COL_XTB);
+				DEL_BIT(vtc, COL_256);
 
 				ses->vtc = vtc;
 				ses->fgc = fgc;
 				ses->bgc = bgc;
 				break;
 
+			case  6:
+				*pto++ = '&';
+				pti++;
+				break;
+
+			case 28:
+				*pto++ = '<';
+				pti++;
+				break;
+
+			case 30:
+				*pto++ = '>';
+				pti++;
+				break;
+
 			case '>':
 				sprintf(pto, "&gt;");
 				pto += strlen(pto);
@@ -411,10 +374,6 @@ void vt102_to_html(struct session *ses, char *txt, char *out)
 				pti++;
 				break;
 
-			case '\r':
-				pti++;
-				break;
-
 			case '"':
 				sprintf(pto, "&quot;");
 				pto += strlen(pto);
