@@ -75,6 +75,24 @@ void winch_handler(int signal)
 
 void abort_handler(int signal)
 {
+	static char crashed = FALSE;
+
+	if (crashed)
+	{
+		exit(-1);
+	}
+	crashed = TRUE;
+
+	reset_terminal();
+
+	clean_screen(gtd->ses);
+
+	dump_stack();
+
+	fflush(NULL);
+
+	exit(-1);
+
 	if (gtd->ses->connect_retry > utime())
 	{
 		gtd->ses->connect_retry = 0;
@@ -135,12 +153,12 @@ void trap_handler(int signal)
 
 int main(int argc, char **argv)
 {
+	int greeting = TRUE;
+
 	#ifdef SOCKS
 		SOCKSinit(argv[0]);
 	#endif
 
-	init_tintin();
-
 	if (signal(SIGTERM, trap_handler) == BADSIG)
 	{
 		syserr("signal SIGTERM");
@@ -198,16 +216,46 @@ int main(int argc, char **argv)
 	{
 		int c;
 
-		while ((c = getopt(argc, argv, "e: h r: t: v")) != EOF)
+		while ((c = getopt(argc, argv, "e: G h r: t: v")) != EOF)
+		{
+			if (c == 'G')
+			{
+				greeting = FALSE;
+			}
+		}
+
+		optind = 1;
+	}
+
+	init_tintin(greeting);
+
+	if (argc > 1)
+	{
+		int c;
+
+		optind = 1;
+
+		while ((c = getopt(argc, argv, "e: G h r: t: v")) != EOF)
 		{
 			switch (c)
 			{
 				case 'e':
-					gtd->ses = pre_parse_input(gtd->ses, optarg, SUB_NONE);
+					gtd->ses = script_driver(gtd->ses, -1, optarg);
+					break;
+
+				case 'G':
 					break;
 
 				case 'h':
-					tintin_printf(NULL, "Usage: %s [-v] [file] [-e \"command\"] [-r file] [-t title]", argv[0]);
+					tintin_printf(NULL, "Usage: %s [OPTION]... [FILE]...", argv[0]);
+					tintin_printf(NULL, "");
+					tintin_printf(NULL, "  -e  Execute given command.");
+					tintin_printf(NULL, "  -G  Don't show the greeting screen.");
+					tintin_printf(NULL, "  -h  This help section.");
+					tintin_printf(NULL, "  -r  Read given file.");
+					tintin_printf(NULL, "  -t  Set given title.");
+					tintin_printf(NULL, "  -v  Enable verbose mode.");
+
 					reset_terminal();
 					exit(1);
 					break;
@@ -229,12 +277,13 @@ int main(int argc, char **argv)
 					break;
 			}
 		}
-		
+
 		if (argv[optind] != NULL)
 		{
 			gtd->ses = do_read(gtd->ses, argv[optind]);
 		}
-	} 
+	}
+	check_all_events(gts, 0, 0, "PROGRAM START");
 
 	mainloop();
 
@@ -242,54 +291,54 @@ int main(int argc, char **argv)
 }
 
 
-void init_tintin(void)
+void init_tintin(int greeting)
 {
-	int ref, cnt;
+	int ref, index;
 
-	gts                 = calloc(1, sizeof(struct session));
+	gts = (struct session *) calloc(1, sizeof(struct session));
 
-	for (cnt = 0 ; cnt < LIST_MAX ; cnt++)
+	for (index = 0 ; index < LIST_MAX ; index++)
 	{
-		gts->list[cnt] = init_list(cnt);
+		gts->list[index] = init_list(gts, index, 32);
 	}
 
 	gts->name           = strdup("gts");
-	gts->class          = strdup("");
+	gts->group          = strdup("");
 	gts->host           = strdup("");
 	gts->port           = strdup("");
-	gts->timestamp      = strdup("");
+	gts->cmd_color      = strdup("");
 	gts->telopts        = TELOPT_FLAG_ECHO;
+	gts->flags          = SES_FLAG_MCCP;
 	gts->socket         = 1;
+	gts->read_max       = 16384;
 
-	gtd                 = calloc(1, sizeof(struct tintin_data));
+	gtd                 = (struct tintin_data *) calloc(1, sizeof(struct tintin_data));
 
 	gtd->ses            = gts;
 
-	gtd->mem            = calloc(1, sizeof(struct memory_data));
-
 	gtd->str_hash_size  = sizeof(struct str_hash_data);
 
-	gtd->mccp_buf_max   = 64;
-	gtd->mccp_buf       = calloc(1, gtd->mccp_buf_max);
+	gtd->mccp_len       = 4096;
+	gtd->mccp_buf       = (unsigned char *) calloc(1, gtd->mccp_len);
 
-	gtd->mud_output_max = 64;
-	gtd->mud_output_buf = calloc(1, gtd->mud_output_max);
+	gtd->mud_output_max = 16384;
+	gtd->mud_output_buf = (char *) calloc(1, gtd->mud_output_max);
 
 	gtd->input_off      = 1;
 
-	for (cnt = 0 ; cnt < 99 ; cnt++)
+	for (index = 0 ; index < 100 ; index++)
 	{
-		gtd->vars[cnt] = strdup("");
-		gtd->cmds[cnt] = strdup("");
+		gtd->vars[index] = strdup("");
+		gtd->cmds[index] = strdup("");
 	}
 
 	for (ref = 0 ; ref < 26 ; ref++)
 	{
-		for (cnt = 0 ; *command_table[cnt].name != 0 ; cnt++)
+		for (index = 0 ; *command_table[index].name != 0 ; index++)
 		{
-			if (*command_table[cnt].name == 'a' + ref)
+			if (*command_table[index].name == 'a' + ref)
 			{
-				gtd->command_ref[ref] = cnt;
+				gtd->command_ref[ref] = index;
 				break;
 			}
 		}
@@ -303,53 +352,44 @@ void init_tintin(void)
 
 	printf("\033=\033[?1h");
 
-	do_configure(gts, "{SPEEDWALK}         {OFF}");
-	do_configure(gts, "{VERBATIM}          {OFF}");
-	do_configure(gts, "{REPEAT ENTER}      {OFF}");
+	do_configure(gts, "{AUTO TAB}         {5000}");
+	do_configure(gts, "{BUFFER SIZE}     {20000}");
 	do_configure(gts, "{COMMAND COLOR}   {<078>}");
 	do_configure(gts, "{COMMAND ECHO}       {ON}");
+	do_configure(gts, "{CONNECT RETRY}      {15}");
+	do_configure(gts, "{HISTORY SIZE}     {1000}");
+	do_configure(gts, "{LOG}               {RAW}");
+	do_configure(gts, "{PACKET PATCH}     {0.01}");
+	do_configure(gts, "{REPEAT CHAR}         {!}");
+	do_configure(gts, "{REPEAT ENTER}      {OFF}");
+	do_configure(gts, "{SCROLL LOCK}        {ON}");
+	do_configure(gts, "{SPEEDWALK}         {OFF}");
+	do_configure(gts, "{TINTIN CHAR}         {#}");
+	do_configure(gts, "{VERBATIM}          {OFF}");
+	do_configure(gts, "{VERBATIM CHAR}      {\\}");
 	do_configure(gts, "{VERBOSE}           {OFF}");
 	do_configure(gts, "{WORDWRAP}           {ON}");
-	do_configure(gts, "{LOG}               {RAW}");
-	do_configure(gts, "{BUFFER SIZE}      {5000}");
-	do_configure(gts, "{SCROLL LOCK}        {ON}");
-	do_configure(gts, "{HISTORY SIZE}     {2500}");
-	do_configure(gts, "{CONNECT RETRY}      {15}");
-	do_configure(gts, "{PACKET PATCH}        {0}");
-	do_configure(gts, "{TINTIN CHAR}         {#}");
-	do_configure(gts, "{VERBATIM CHAR}      {\\}");
-	do_configure(gts, "{REPEAT CHAR}         {!}");
 
-	insertnode_list(gts,  "n",  "s",  "1", LIST_PATHDIR);
-	insertnode_list(gts,  "e",  "w",  "2", LIST_PATHDIR);
-	insertnode_list(gts,  "s",  "n",  "4", LIST_PATHDIR);
-	insertnode_list(gts,  "w",  "e",  "8", LIST_PATHDIR);
-	insertnode_list(gts,  "u",  "d", "16", LIST_PATHDIR);
-	insertnode_list(gts,  "d",  "u", "32", LIST_PATHDIR);
+	insert_node_list(gts->list[LIST_PATHDIR],  "n",  "s",  "1");
+	insert_node_list(gts->list[LIST_PATHDIR],  "e",  "w",  "2");
+	insert_node_list(gts->list[LIST_PATHDIR],  "s",  "n",  "4");
+	insert_node_list(gts->list[LIST_PATHDIR],  "w",  "e",  "8");
+	insert_node_list(gts->list[LIST_PATHDIR],  "u",  "d", "16");
+	insert_node_list(gts->list[LIST_PATHDIR],  "d",  "u", "32");
 
-	insertnode_list(gts, "ne", "sw",  "3", LIST_PATHDIR);
-	insertnode_list(gts, "nw", "se",  "9", LIST_PATHDIR);
-	insertnode_list(gts, "se", "nw",  "6", LIST_PATHDIR);
-	insertnode_list(gts, "sw", "ne", "12", LIST_PATHDIR);
+	insert_node_list(gts->list[LIST_PATHDIR], "ne", "sw",  "3");
+	insert_node_list(gts->list[LIST_PATHDIR], "nw", "se",  "9");
+	insert_node_list(gts->list[LIST_PATHDIR], "se", "nw",  "6");
+	insert_node_list(gts->list[LIST_PATHDIR], "sw", "ne", "12");
 
 	init_terminal();
 
-	do_showme(gts, "");
-	do_showme(gts, "<068>      #<068>###################################################################<068>#");
-	do_showme(gts, "<068>      #<078>                                                                   <068>#");
-	do_showme(gts, "<068>      #<078>                           T I N T I N + +                         <068>#");
-	do_showme(gts, "<068>      #<078>                                                                   <068>#");
-	do_showme(gts, "<068>      #<078>            (<068>T<078>)he k(<068>I<078>)cki(<068>N<078>) (<068>T<078>)ickin d(<068>I<078>)kumud clie(<068>N<078>)t           <068>#");
-	do_showme(gts, "<068>      #<078>                                                                   <068>#");
-	do_showme(gts, "<068>      #<078>                 Original TINTIN code by Peter Unold               <068>#");
-	do_showme(gts, "<068>      #<078>       new code by Bill Reiss, David A. Wagner, Rob Ellsworth,     <068>#");
-	do_showme(gts, "<068>      #<078>                 Jeremy C. Jack, Igor van den Hoven                <068>#");
-	do_showme(gts, "<068>      #<078>                             1992, 2007                            <068>#");
-	do_showme(gts, "<068>      #<078>                                                                   <068>#");
-	do_showme(gts, "<068>      #<078>                               "VERSION_NUM"                              <068>#");
-	do_showme(gts, "<068>      #<078>                                                                   <068>#");
-	do_showme(gts, "<068>      #<068>###################################################################<068>#<088>");
-	do_showme(gts, "");
+	if (greeting)
+	{
+		do_advertise(gts, "");
+
+		do_help(gts, "GREETING");
+	}
 }
 
 
@@ -357,6 +397,8 @@ void quitmsg(char *message)
 {
 	struct session *ses;
 
+	SET_BIT(gtd->flags, TINTIN_FLAG_TERMINATE);
+
 	while ((ses = gts->next) != NULL)
 	{
 		cleanup_session(ses);
@@ -367,7 +409,8 @@ void quitmsg(char *message)
 		close(gtd->chat->fd);
 	}
 
-	check_all_events(gts, "PROGRAM TERMINATION");
+	check_all_events(gts, 0, 0, "PROGRAM TERMINATION");
+
 /*
 	if (gtd->history_size)
 	{
@@ -375,7 +418,7 @@ void quitmsg(char *message)
 
 		sprintf(filename, "%s/%s", getenv("HOME"), HISTORY_FILE);
 
-		write_history(gts, filename);
+		history_write(gts, filename);
 	}
 */
 	reset_terminal();
