@@ -37,11 +37,13 @@ DO_COMMAND(do_all)
 	{
 		get_arg_in_braces(arg, left, TRUE);
 
+		substitute(ses, left, left, SUB_VAR|SUB_FUN);
+
 		for (sesptr = gts->next ; sesptr ; sesptr = next_ses)
 		{
 			next_ses = sesptr->next;
 
-			pre_parse_input(sesptr, left, SUB_NONE);
+			script_driver(sesptr, -1, left);
 		}
 	}
 	else
@@ -73,7 +75,7 @@ DO_COMMAND(do_commands)
 		{
 			continue;
 		}
-		if (strlen(buf) + 20 > ses->cols)
+		if ((int) strlen(buf) + 20 > ses->cols)
 		{
 			tintin_puts2(ses, buf);
 			buf[0] = 0;
@@ -101,26 +103,50 @@ DO_COMMAND(do_cr)
 
 DO_COMMAND(do_echo)
 {
-	char temp[BUFFER_SIZE], output[BUFFER_SIZE];
+	char temp[BUFFER_SIZE], output[BUFFER_SIZE], left[BUFFER_SIZE];
+	int lnf;
 
-	sprintf(temp, "result %s", arg);
+	sprintf(temp, "{result} %s", arg);
 
-	internal_variable(ses, "{result}");
+	set_nest_node(ses->list[LIST_VARIABLE], "result", "");
 
 	do_format(ses, temp);
 
-	substitute(ses, "$result", temp, SUB_VAR|SUB_FUN|SUB_COL|SUB_ESC);
+	substitute(ses, "$result", temp, SUB_VAR|SUB_FUN|SUB_COL);
+
+	arg = temp;
+
+	if (*arg == DEFAULT_OPEN)
+	{
+		arg = get_arg_in_braces(arg, left, TRUE);
+		      get_arg_in_braces(arg, temp, TRUE);
+
+		if (*temp)
+		{
+			int row = (int) get_number(ses, temp);
+
+			substitute(ses, left, temp, SUB_ESC);
+
+			do_one_prompt(ses, temp, row);
+
+			return ses;
+		}
+	}
+
+	lnf = !str_suffix(arg, "\\");
+
+	substitute(ses, arg, temp, SUB_ESC);
 
 	if (strip_vt102_strlen(ses->more_output) != 0)
 	{
-		sprintf(output, "\r\n\033[0m%s\033[0m", temp);
+		sprintf(output, "\n\033[0m%s\033[0m", temp);
 	}
 	else
 	{
 		sprintf(output, "\033[0m%s\033[0m", temp);
 	}
 
-	add_line_buffer(ses, output, FALSE);
+	add_line_buffer(ses, output, lnf);
 
 	if (ses != gtd->ses)
 	{
@@ -133,7 +159,7 @@ DO_COMMAND(do_echo)
 		goto_rowcol(ses, ses->bot_row, 1);
 	}
 
-	printline(ses, output, FALSE);
+	printline(ses, output, lnf);
 
 	if (!HAS_BIT(ses->flags, SES_FLAG_READMUD) && IS_SPLIT(ses))
 	{
@@ -161,11 +187,9 @@ DO_COMMAND(do_forall)
 {
 	char left[BUFFER_SIZE], right[BUFFER_SIZE], temp[BUFFER_SIZE];
 
-	arg = get_arg_in_braces(arg, left,  TRUE);
+	arg = sub_arg_in_braces(ses, arg, left, GET_ALL, SUB_VAR|SUB_FUN);
 	arg = get_arg_in_braces(arg, right, TRUE);
 
-	substitute(ses, left, left, SUB_VAR|SUB_FUN);
-
 	if (*left == 0 || *right == 0)
 	{
 		tintin_printf2(ses, "#ERROR: #FORALL - PROVIDE 2 ARGUMENTS");
@@ -176,32 +200,28 @@ DO_COMMAND(do_forall)
 
 		while (*arg)
 		{
-			arg = get_arg_in_braces(arg, temp, FALSE);
-
-			internal_variable(ses, "{forall} {%s}", temp);
+			arg = get_arg_in_braces(arg, temp, TRUE);
 
 			RESTRING(gtd->cmds[0], temp);
 
-			ses = pre_parse_input(ses, right, SUB_CMD);
+			substitute(ses, right, temp, SUB_CMD);
+
+			ses = script_driver(ses, -1, temp);
+
+			if (*arg == COMMAND_SEPARATOR)
+			{
+				arg++;
+			}
 		}
 	}
 	return ses;
 }
 
-DO_COMMAND(do_gagline)
-{
-	SET_BIT(ses->flags, SES_FLAG_GAG);
-
-	return ses;
-}
-
 
 DO_COMMAND(do_info)
 {
 	int cnt;
 
-	printf("--%s--\n", arg);
-
 	if (*arg == 'c')
 	{
 		show_cpu(ses);
@@ -209,6 +229,13 @@ DO_COMMAND(do_info)
 		return ses;
 	}
 
+	if (*arg == 's')
+	{
+		dump_stack();
+
+		return ses;
+	}
+
 	tintin_header(ses, " INFORMATION ");
 
 	for (cnt = 0 ; cnt < LIST_MAX ; cnt++)
@@ -219,7 +246,7 @@ DO_COMMAND(do_info)
 		}
 		tintin_printf2(ses, "%-20s  %5d  IGNORE %3s  MESSAGE %3s  DEBUG %3s",
 			list_table[cnt].name_multi,
-			ses->list[cnt]->count,
+			ses->list[cnt]->used,
 			HAS_BIT(ses->list[cnt]->flags, LIST_FLAG_IGNORE) ? "ON" : "OFF",
 			HAS_BIT(ses->list[cnt]->flags, LIST_FLAG_MESSAGE) ? "ON" : "OFF",
 			HAS_BIT(ses->list[cnt]->flags, LIST_FLAG_DEBUG)   ? "ON" : "OFF");
@@ -230,62 +257,6 @@ DO_COMMAND(do_info)
 }
 
 
-DO_COMMAND(do_loop)
-{
-	char left[BUFFER_SIZE], right[BUFFER_SIZE], temp[BUFFER_SIZE], command[BUFFER_SIZE];
-
-	long long bound1, bound2, counter, step;
-
-	arg = get_arg_in_braces(arg, temp,    0);
-	arg = get_arg_in_braces(arg, command, 1);
-
-	arg = get_arg_in_braces(temp, left, 0);
-	arg = get_arg_in_braces(arg, right, 0);
-
-
-	if (*command == 0 || (*left == 0 && *right == 0))
-	{
-		show_message(ses, LIST_MATH, "#LOOP: INVALID ARGUMENTS.");
-
-		return ses;
-	}
-
-	if (*left && *right)
-	{
-		bound1 = (long long) get_number(ses, left);
-		bound2 = (long long) get_number(ses, right);
-
-		if (bound1 <= bound2)
-		{
-			step = 1;
-			bound2++;
-		}
-		else
-		{
-			step = -1;
-			bound2--;
-		}
-
-		for (counter = bound1 ; counter != bound2 ; counter += step)
-		{
-			gtd->cmds[0] = refstring(gtd->cmds[0], "%lld", counter);
-
-			internal_variable(ses, "{loop} {%lld}", counter);
-
-			ses = pre_parse_input(ses, command, SUB_CMD);
-		}
-	}
-	else
-	{
-		while ((long long) get_number(ses, left))
-		{
-			ses = pre_parse_input(ses, command, SUB_NONE);
-		}
-	}
-	return ses;
-}
-
-
 DO_COMMAND(do_nop)
 {
 	return ses;
@@ -327,53 +298,43 @@ DO_COMMAND(do_parse)
 
 			RESTRING(gtd->cmds[0], temp);
 
-			internal_variable(ses, "{parse} {%s}", gtd->cmds[0]);
+			substitute(ses, right, temp, SUB_CMD);
 
-			ses = pre_parse_input(ses, right, SUB_CMD);
+			ses = script_driver(ses, -1, temp);
 		}
 	}
 	return ses;
 }
 
 
-DO_COMMAND(do_return)
-{
-	SET_BIT(ses->flags, SES_FLAG_BREAK);
-
-	if (*arg)
-	{
-		internal_variable(ses, "{result} %s", arg);
-	}
-	return ses;
-}
-
 DO_COMMAND(do_send)
 {
 	char left[BUFFER_SIZE];
 
+	push_call("do_send(%p,%p)",ses,arg);
+
 	get_arg_in_braces(arg, left, TRUE);
 
 	write_mud(ses, left, SUB_VAR|SUB_FUN|SUB_ESC|SUB_EOL);
 
+	pop_call();
 	return ses;
 }
 
 DO_COMMAND(do_showme)
 {
-	char left[BUFFER_SIZE], right[BUFFER_SIZE], output[STRING_SIZE];
+	char left[BUFFER_SIZE], right[BUFFER_SIZE], temp[STRING_SIZE];
+	int lnf;
 
-	arg = get_arg_in_braces(arg, left,  1);
-	arg = get_arg_in_braces(arg, right, 0);
+	arg = get_arg_in_braces(arg, left, TRUE);
 
-	substitute(ses, left, left, SUB_COL|SUB_ESC);
+	lnf = !str_suffix(left, "\\");
 
-	if (*right)
-	{
-		do_one_prompt(ses, left, atoi(right));
+	substitute(ses, left, temp, SUB_VAR|SUB_FUN|SUB_COL|SUB_ESC);
 
-		return ses;
-	}
-	do_one_line(left, ses);
+	arg = sub_arg_in_braces(ses, arg, right, GET_ONE, SUB_VAR|SUB_FUN);
+
+	do_one_line(temp, ses);
 
 	if (HAS_BIT(ses->flags, SES_FLAG_GAG))
 	{
@@ -382,16 +343,23 @@ DO_COMMAND(do_showme)
 		return ses;
 	}
 
+	if (*right)
+	{
+		do_one_prompt(ses, temp, (int) get_number(ses, right));
+
+		return ses;
+	}
+
 	if (strip_vt102_strlen(ses->more_output) != 0)
 	{
-		sprintf(output, "\r\n\033[0m%s\033[0m", left);
+		sprintf(right, "\n\033[0m%s\033[0m", temp);
 	}
 	else
 	{
-		sprintf(output, "\033[0m%s\033[0m", left);
+		sprintf(right, "\033[0m%s\033[0m", temp);
 	}
 
-	add_line_buffer(ses, output, FALSE);
+	add_line_buffer(ses, right, lnf);
 
 	if (ses != gtd->ses)
 	{
@@ -404,7 +372,7 @@ DO_COMMAND(do_showme)
 		goto_rowcol(ses, ses->bot_row, 1);
 	}
 
-	printline(ses, output, FALSE);
+	printline(ses, right, lnf);
 
 	if (!HAS_BIT(ses->flags, SES_FLAG_READMUD) && IS_SPLIT(ses))
 	{
@@ -420,6 +388,7 @@ DO_COMMAND(do_snoop)
 	char left[BUFFER_SIZE];
 
 	get_arg_in_braces(arg, left, 1);
+	substitute(ses, left, left, SUB_VAR|SUB_FUN);
 
 	if (*left)
 	{
@@ -463,6 +432,19 @@ DO_COMMAND(do_suspend)
 }
 
 
+DO_COMMAND(do_test)
+{
+	char buf[BUFFER_SIZE];
+
+	strcpy(buf, get_alnum(ses, arg));
+
+	tintin_printf2(ses, "is_number: %d", is_number(buf));
+
+	tintin_printf2(ses, "%s", get_alnum(ses, buf));
+
+	return ses;
+}
+
 DO_COMMAND(do_zap)
 {
 	tintin_puts(ses, "");
