@@ -46,7 +46,9 @@ DO_COMMAND(do_run)
 	char *argv[4] = {"sh", "-c", "", NULL};
 
 	arg = get_arg_in_braces(arg, left,  FALSE);
+	substitute(ses, left, left, SUB_VAR|SUB_FUN);
 	arg = get_arg_in_braces(arg, right, TRUE);
+	substitute(ses, right, right, SUB_VAR|SUB_FUN);
 
 	if (*left == 0 || *right == 0)
 	{
@@ -71,7 +73,7 @@ DO_COMMAND(do_run)
 			break;
 
 		default:
-			ses = new_session(ses, left, "", desc);
+			ses = new_session(ses, left, right, desc);
 			break;
 	}
 	return gtd->ses;
@@ -83,6 +85,7 @@ DO_COMMAND(do_scan)
 	char filename[BUFFER_SIZE];
 
 	get_arg_in_braces(arg, filename, TRUE);
+	substitute(ses, filename, filename, SUB_VAR|SUB_FUN);
 
 	if ((fp = fopen(filename, "r")) == NULL)
 	{
@@ -95,7 +98,7 @@ DO_COMMAND(do_scan)
 	if (STRING_SIZE > gtd->mud_output_max)
 	{
 		gtd->mud_output_max  = STRING_SIZE;
-		gtd->mud_output_buf  = realloc(gtd->mud_output_buf, gtd->mud_output_max);
+		gtd->mud_output_buf  = (char *) realloc(gtd->mud_output_buf, gtd->mud_output_max);
 	}
 
 	while (fgets(gtd->mud_output_buf, STRING_SIZE, fp))
@@ -116,20 +119,20 @@ DO_COMMAND(do_scan)
 
 DO_COMMAND(do_script)
 {
-	char left[BUFFER_SIZE], right[BUFFER_SIZE], *cptr, buffer[BUFFER_SIZE];
-	struct listnode *node;
+	char arg1[BUFFER_SIZE], arg2[BUFFER_SIZE], *cptr, buffer[BUFFER_SIZE], var[BUFFER_SIZE];
 	FILE *script;
+	int index;
 
-	arg = get_arg_in_braces(arg, left, TRUE);
-	arg = get_arg_in_braces(arg, right, TRUE);
+	arg = sub_arg_in_braces(ses, arg, arg1, GET_ALL, SUB_VAR|SUB_FUN);
+	arg = sub_arg_in_braces(ses, arg, arg2, GET_ALL, SUB_VAR|SUB_FUN);
 
-	if (*left == 0)
+	if (*arg1 == 0)
 	{
 		show_message(ses, LIST_MESSAGE, "#SCRIPT: ONE ARGUMENT REQUIRED.");
 	}
-	else if (*right == 0)
+	else if (*arg2 == 0)
 	{
-		script = popen(left, "r");
+		script = popen(arg1, "r");
 
 		while (fgets(buffer, BUFFER_SIZE - 1, script))
 		{
@@ -140,18 +143,18 @@ DO_COMMAND(do_script)
 				*cptr = 0;
 			}
 
-			pre_parse_input(ses, buffer, SUB_NONE);
+			ses = script_driver(ses, -1, buffer);
 		}
 
 		pclose(script);
 	}
 	else
 	{
-		script = popen(right, "r");
+		index = 1;
 
-		internal_variable(ses, "{%s}", left);
+		script = popen(arg2, "r");
 
-		node = searchnode_list(ses->list[LIST_VARIABLE], left);
+		var[0] = 0;
 
 		while (fgets(buffer, BUFFER_SIZE - 1, script))
 		{
@@ -161,8 +164,9 @@ DO_COMMAND(do_script)
 			{
 				*cptr = 0;
 			}
-			array_ins(ses, node, "-1", buffer);
+			cat_sprintf(var, "{%d}{%s}", index++, buffer);
 		}
+		set_nest_node(ses->list[LIST_VARIABLE], arg1, "%s", var);
 
 		pclose(script);
 	}
@@ -174,6 +178,7 @@ DO_COMMAND(do_system)
 	char left[BUFFER_SIZE];
 
 	get_arg_in_braces(arg, left, TRUE);
+	substitute(ses, left, left, SUB_VAR|SUB_FUN);
 
 	if (*left == 0)
 	{
@@ -183,20 +188,20 @@ DO_COMMAND(do_system)
 
 	show_message(ses, LIST_MESSAGE, "#OK: EXECUTING '%s'", left);
 
-	if (!HAS_BIT(ses->flags, SES_FLAG_READMUD) && IS_SPLIT(ses))
+	if (!HAS_BIT(gtd->ses->flags, SES_FLAG_READMUD) && IS_SPLIT(gtd->ses))
 	{
-		save_pos(ses);
-		goto_rowcol(ses, ses->bot_row, 1);
+		save_pos(gtd->ses);
+		goto_rowcol(gtd->ses, gtd->ses->bot_row, 1);
 	}
-
 	fflush(stdout);
 
 	system(left);
 
-	if (!HAS_BIT(ses->flags, SES_FLAG_READMUD) && IS_SPLIT(ses))
+	if (!HAS_BIT(gtd->ses->flags, SES_FLAG_READMUD) && IS_SPLIT(gtd->ses))
 	{
-		restore_pos(ses);
+		restore_pos(gtd->ses);
 	}
+	fflush(stdout);
 
 	return ses;
 }
@@ -205,13 +210,16 @@ DO_COMMAND(do_system)
 DO_COMMAND(do_textin)
 {
 	FILE *fp;
-	char filename[BUFFER_SIZE], buffer[BUFFER_SIZE], *cptr;
+	char left[BUFFER_SIZE], right[BUFFER_SIZE], buffer[BUFFER_SIZE], *cptr;
 
-	get_arg_in_braces(arg, filename, TRUE);
+	arg = get_arg_in_braces(arg, left, FALSE);
+	substitute(ses, left, left, SUB_VAR|SUB_FUN);
 
-	if ((fp = fopen(filename, "r")) == NULL)
+	arg = get_arg_in_braces(arg, right, TRUE);
+
+	if ((fp = fopen(left, "r")) == NULL)
 	{
-		tintin_printf(ses, "#ERROR: #TEXTIN {%s} - FILE NOT FOUND.", filename);
+		tintin_printf(ses, "#ERROR: #TEXTIN {%s} - FILE NOT FOUND.", left);
 		return ses;
 	}
 
@@ -225,6 +233,11 @@ DO_COMMAND(do_textin)
 		}
 
 		write_mud(ses, buffer, SUB_EOL);
+
+		if (*right)
+		{
+			usleep((long long) (get_number(ses, right) * 1000000));
+		}
 	}
 	fclose(fp);
 
