@@ -26,256 +26,179 @@
 *                     recoded by Igor van den Hoven 2004                      *
 ******************************************************************************/
 
-#include <regex.h>
 #include <sys/types.h>
+#include <pcre.h>
 
 #include "tintin.h"
 
 
-/******************************************************************************
-* Simple regexp that skips vt102 codes and supports the * and ? wildcards     *
-******************************************************************************/
-
 int match(struct session *ses, char *str, char *exp)
 {
-	char expbuf[BUFFER_SIZE], tmpbuf[BUFFER_SIZE];
+	char expbuf[BUFFER_SIZE];
 
-	if (HAS_BIT(ses->flags, SES_FLAG_REGEXP))
+	if (ses)
 	{
-		substitute(ses, exp, tmpbuf, SUB_VAR|SUB_FUN|SUB_ESC);
+		sprintf(expbuf, "^%s$", exp);
 
-		sprintf(expbuf, "^%s$", tmpbuf);
-
-		return regexp(str, expbuf);
+		substitute(ses, expbuf, expbuf, SUB_VAR|SUB_FUN);
 	}
 	else
 	{
+		sprintf(expbuf, "^%s$", exp);
+	}
+	return tintin_regexp(NULL, str, expbuf, 0, 0);
+}
+
+int find(struct session *ses, char *str, char *exp, int sub)
+{
+	char expbuf[BUFFER_SIZE], strbuf[BUFFER_SIZE];
+
+	if (ses)
+	{
+		substitute(ses, str, strbuf, SUB_VAR|SUB_FUN);
 		substitute(ses, exp, expbuf, SUB_VAR|SUB_FUN);
 
-		return glob(str, expbuf);
-	}
-}
-
-int find(struct session *ses, char *str, char *exp)
-{
-	char expbuf[BUFFER_SIZE], tmpbuf[BUFFER_SIZE];
-
-	if (HAS_BIT(ses->flags, SES_FLAG_REGEXP))
-	{
-		strip_vt102_codes(str, tmpbuf);
-
-		substitute(ses, exp, expbuf, SUB_VAR|SUB_FUN|SUB_ESC);
-
-		return regexp(tmpbuf, expbuf);
+		return tintin_regexp(NULL, strbuf, expbuf, 0, sub);
 	}
 	else
 	{
-		substitute(ses, exp, tmpbuf, SUB_VAR|SUB_FUN);
-
-		sprintf(expbuf, "*%s*", tmpbuf);
-
-		return glob(str, expbuf);
+		return tintin_regexp(NULL, str, exp, 0, sub);
 	}
 }
 
-int glob(char *str, char *exp)
-{
-	short cnt;
-
-	while (*exp)
-	{
-		while ((cnt = skip_vt102_codes(str)) != 0)
-		{
-			str += cnt;
-		}
-
-#ifdef BIG5
-		if (*exp & 0x80 && *str & 0x80)
-		{
-			if (*++exp != *++str)
-			{
-				return FALSE;
-			}
-			continue;
-		}
-#endif
-
-		switch (*exp)
-		{
-			case '?':
-				if (*str == 0)
-				{
-					return FALSE;
-				}
-				exp++;
-				str++;
-				break;
-
-			case '*':
-				for (cnt = strlen(str) ; cnt >= 0 ; cnt--)
-				{
-					if (glob(&str[cnt], exp + 1))
-					{
-						return TRUE;
-					}
-				}
-				return FALSE;
-
-			case '\\':
-				exp++;
-
-			default:
-				if (*exp++ != *str++)
-				{
-					return FALSE;
-				}
-				break;
-		}
-	}
-	return (*exp == *str);
-}
-
 DO_COMMAND(do_regexp)
 {
-	char left[BUFFER_SIZE], right[BUFFER_SIZE], true[BUFFER_SIZE], false[BUFFER_SIZE];
+	char left[BUFFER_SIZE], right[BUFFER_SIZE], is_true[BUFFER_SIZE], is_false[BUFFER_SIZE];
 
 	arg = get_arg_in_braces(arg, left,  0);
 	arg = get_arg_in_braces(arg, right, 0);
-	arg = get_arg_in_braces(arg, true,  1);
-	arg = get_arg_in_braces(arg, false, 1);
+	arg = get_arg_in_braces(arg, is_true,  1);
+	arg = get_arg_in_braces(arg, is_false, 1);
 
-	if (*left == 0 || *right == 0 || *true == 0)
+	if (*is_true == 0)
 	{
 		tintin_printf2(ses, "SYNTAX: #REGEXP {string} {expression} {true} {false}.");
 	}
 	else
 	{
 		substitute(ses, left,  left,  SUB_VAR|SUB_FUN);
-		substitute(ses, right, right, SUB_VAR|SUB_FUN|SUB_ESC);
+		substitute(ses, right, right, SUB_VAR|SUB_FUN);
 
-		if (regexp_cmds(left, right))
+		if (tintin_regexp(NULL, left, right, 0, SUB_CMD))
 		{
-			ses = pre_parse_input(ses, true, SUB_CMD);
+			substitute(ses, is_true, is_true, SUB_CMD);
+
+			ses = script_driver(ses, -1, is_true);
 		}
-		else if (*false)
+		else if (*is_false)
 		{
-			ses = pre_parse_input(ses, false, SUB_NONE);
+			ses = script_driver(ses, -1, is_false);
 		}
 	}
 	return ses;
 }
 
-int regexp_cmds(char *str, char *exp)
+int regexp_compare(pcre *nodepcre, char *str, char *exp, int option, int flag)
 {
-	regex_t regex;
-	regmatch_t match[100];
-	int i;
+	pcre *regex;
+	const char *error;
+	int i, j, matches, match[303];
 
-	if (regcomp(&regex, exp, REG_EXTENDED))
+	if (nodepcre == NULL)
 	{
-		return FALSE;
-	}
-
-	if (regexec(&regex, str, 100, match, 0))
-	{
-		regfree(&regex);
-
-		return FALSE;
-	}
-
-	for (i = 0 ; i < 100 ; i++)
-	{
-		if (match[i].rm_so != -1)
-		{
-			gtd->cmds[i] = refstring(gtd->cmds[i], "%.*s", match[i].rm_eo - match[i].rm_so, &str[match[i].rm_so]);
-		}
-		else
-		{
-			break;
-		}
-	}
-	regfree(&regex);
-
-	return TRUE;
-}
-
-int action_regexp(char *str, char *exp)
-{
-	regex_t regex;
-	regmatch_t match[100];
-	int i;
-
-	if (regcomp(&regex, exp, REG_EXTENDED))
-	{
-		return FALSE;
-	}
-
-	if (regexec(&regex, str, 100, match, 0))
-	{
-		regfree(&regex);
-
-		return FALSE;
-	}
-
-	for (i = 0 ; i < 100 ; i++)
-	{
-		if (match[i].rm_so == -1)
-		{
-			RESTRING(gtd->vars[i], "");
-		}
-		else
-		{
-			gtd->vars[i] = refstring(gtd->vars[i], "%.*s", match[i].rm_eo - match[i].rm_so, &str[match[i].rm_so]);
-		}
-	}
-
-	regfree(&regex);
-
-	return TRUE;
-}
-
-int regexp(char *str, char *exp)
-{
-	regex_t regex;
-	regmatch_t match[1];
-
-	if (regcomp(&regex, exp, REG_NOSUB|REG_EXTENDED))
-	{
-		return FALSE;
-	}
-
-	if (regexec(&regex, str, 0, match, 0))
-	{
-		regfree(&regex);
-
-		return FALSE;
+		regex = pcre_compile(exp, option, &error, &i, NULL);
 	}
 	else
 	{
-		regfree(&regex);
-
-		return TRUE;
+		regex = nodepcre;
 	}
+
+	if (regex == NULL)
+	{
+		return FALSE;
+	}
+
+	matches = pcre_exec(regex, NULL, str, strlen(str), 0, 0, match, 303);
+
+	if (matches <= 0)
+	{
+		if (nodepcre == NULL)
+		{
+			free(regex);
+		}
+		return FALSE;
+	}
+
+	switch (flag)
+	{
+		case SUB_CMD:
+			for (i = 0 ; i < matches ; i++)
+			{
+				gtd->cmds[i] = refstring(gtd->cmds[i], "%.*s", match[i*2+1] - match[i*2], &str[match[i*2]]);
+			}
+			break;
+
+		case SUB_CMD + SUB_FIX:
+			for (i = 0 ; i < matches ; i++)
+			{
+				j = gtd->args[i];
+
+				gtd->cmds[j] = refstring(gtd->cmds[j], "%.*s", match[i*2+1] - match[i*2], &str[match[i*2]]);
+			}
+			break;
+
+		case SUB_ARG:
+			for (i = 0 ; i < matches ; i++)
+			{
+				gtd->vars[i] = refstring(gtd->vars[i], "%.*s", match[i*2+1] - match[i*2], &str[match[i*2]]);
+			}
+			break;
+
+		case SUB_ARG + SUB_FIX:
+			for (i = 0 ; i < matches ; i++)
+			{
+				j = gtd->args[i];
+
+				gtd->vars[j] = refstring(gtd->vars[j], "%.*s", match[i*2+1] - match[i*2], &str[match[i*2]]);
+			}
+			break;
+	}
+
+	if (nodepcre == NULL)
+	{
+		free(regex);
+	}
+
+	return TRUE;
 }
 
+pcre *regexp_compile(char *exp, int option)
+{
+	const char *error;
+	int i;
+
+	return pcre_compile(exp, option, &error, &i, NULL);
+}
+
+
 /******************************************************************************
 * copy *string into *result, but substitute the various expressions with the  *
 * values they stand for.                                                      *
 ******************************************************************************/
 
-void substitute(struct session *ses, char *string, char *result, int flags)
+int substitute(struct session *ses, char *string, char *result, int flags)
 {
 	struct listnode *node;
 	char temp[BUFFER_SIZE], buf[BUFFER_SIZE], buffer[BUFFER_SIZE], *pti, *pto, *ptt;
 	char *pte;
-	int i, cnt, flags_neol = flags;
+	int i, cnt, escape = FALSE, flags_neol = flags;
 
 	push_call("substitute(%p,%p,%p,%d)",ses,string,result,flags);
 
 	pti = string;
-	pto = buffer;
+	pto = (string == result) ? buffer : result;
 
-	DEL_BIT(flags_neol, SUB_EOL);
+	DEL_BIT(flags_neol, SUB_EOL|SUB_LNF);
 
 	while (TRUE)
 	{
@@ -297,95 +220,88 @@ void substitute(struct session *ses, char *string, char *result, int flags)
 			case '\0':
 				if (HAS_BIT(flags, SUB_EOL))
 				{
-					*pto++ = '\r';
+					if (HAS_BIT(ses->flags, SES_FLAG_RUN))
+					{
+						*pto++ = '\r';
+					}
+					else
+					{
+						*pto++ = '\r';
+						*pto++ = '\n';
+					}
+				}
+
+				if (HAS_BIT(flags, SUB_LNF))
+				{
 					*pto++ = '\n';
 				}
+
 				*pto = 0;
 
-				strcpy(result, buffer);
-
 				pop_call();
-				return;
+
+				if (string == result)
+				{
+					strcpy(result, buffer);
+
+					return pto - buffer;
+				}
+				else
+				{
+					return pto - result;
+				}
 
 			case '$':
 				if (HAS_BIT(flags, SUB_VAR) && (pti[1] == DEFAULT_OPEN || isalnum(pti[1]) || pti[1] == '$'))
 				{
+					int def = FALSE;
+
 					if (pti[1] == '$')
 					{
 						while (pti[1] == '$')
 						{
 							*pto++ = *pti++;
 						}
-						pti++;
 
-						continue;
-					}
-
-					if (pti[1] == DEFAULT_OPEN)
-					{
-						for (ptt = buf, i = 2 ; pti[i] && pti[i] != DEFAULT_CLOSE ; i++)
+						if (pti[1] == DEFAULT_OPEN || isalnum(pti[1]))
 						{
-							*ptt++ = pti[i];
-						}
-
-						if (pti[i] == DEFAULT_CLOSE)
-						{
-							*ptt = 0;
-							substitute(ses, buf, temp, flags_neol);
+							pti++;
 						}
 						else
 						{
-							strcpy(temp, "");
+							*pto++ = *pti++;
 						}
-						i++;
+						continue;
 					}
 
-					if (isalnum(pti[1]) || pti[1] == '_')
+					pti++;
+
+					if (*pti == DEFAULT_OPEN)
 					{
-						for (ptt = temp, i = 1 ; isalnum(pti[i]) || pti[i] == '_' ; i++)
+						def = TRUE;
+
+						pti = get_arg_in_braces(pti, buf, TRUE);
+
+						substitute(ses, buf, temp, flags_neol);
+					}
+					else
+					{
+						ptt = temp;
+
+						while (isalnum(*pti) || *pti == '_')
 						{
-							*ptt++ = pti[i];
+							*ptt++ = *pti++;
 						}
 						*ptt = 0;
 					}
 
-					if ((node = searchnode_list(ses->list[LIST_VARIABLE], temp)) != NULL)
-					{
-						substitute(ses, node->right, pto, flags_neol - SUB_VAR);
+					pti = get_arg_at_brackets(pti, temp + strlen(temp));
 
-						if (pti[1] == DEFAULT_OPEN)
-						{
-							pti += 3 + strlen(buf);
-						}
-						else
-						{
-							pti += 1 + strlen(temp);
-						}
+					get_nest_node(ses->list[LIST_VARIABLE], temp, buf, def);
 
-						if (*pti == '[')
-						{
-							for (ptt = temp, i = 1 ; pti[i] && pti[i] != ']' ; i++)
-							{
-								*ptt++ = pti[i];
-							}
-							*ptt = 0;
+					substitute(ses, buf, pto, flags_neol - SUB_VAR);
 
-							if (pti[i] == ']')
-							{
-								pti += 2 + strlen(temp);
-
-								substitute(ses, temp, temp, SUB_VAR|SUB_FUN);
-
-								get_list_item(ses, node, temp);
-								strcpy(pto, temp);
-							}
-						}
-						pto += strlen(pto);
-					}
-					else
-					{
-						*pto++ = *pti++;
-					}
+					pto += strlen(pto);
 				}
 				else
 				{
@@ -466,7 +382,39 @@ void substitute(struct session *ses, char *string, char *result, int flags)
 						*pto++ = ';';
 						*pto++ = '5';
 						*pto++ = ';';
-						cnt = 16 + (pti[1] - 'A') + (pti[2] - 'A') * 6 + (pti[3] - 'A') * 36;
+						cnt = 16 + (pti[1] - 'A') * 36 + (pti[2] - 'A') * 6 + (pti[3] - 'A');
+						*pto++ = '0' + cnt / 100;
+						*pto++ = '0' + cnt % 100 / 10;
+						*pto++ = '0' + cnt % 10;
+						*pto++ = 'm';
+						pti += 5;
+					}
+					else if (pti[1] == 'g' && isdigit(pti[2]) && isdigit(pti[3]) && pti[4] == '>')
+					{
+						*pto++ = ESCAPE;
+						*pto++ = '[';
+						*pto++ = '3';
+						*pto++ = '8';
+						*pto++ = ';';
+						*pto++ = '5';
+						*pto++ = ';';
+						cnt = 232 + (pti[2] - '0') * 10 + (pti[3] - '0');
+						*pto++ = '0' + cnt / 100;
+						*pto++ = '0' + cnt % 100 / 10;
+						*pto++ = '0' + cnt % 10;
+						*pto++ = 'm';
+						pti += 5;
+					}
+					else if (pti[1] == 'G' && isdigit(pti[2]) && isdigit(pti[3]) && pti[4] == '>')
+					{
+						*pto++ = ESCAPE;
+						*pto++ = '[';
+						*pto++ = '4';
+						*pto++ = '8';
+						*pto++ = ';';
+						*pto++ = '5';
+						*pto++ = ';';
+						cnt = 232 + (pti[2] - '0') * 10 + (pti[3] - '0');
 						*pto++ = '0' + cnt / 100;
 						*pto++ = '0' + cnt % 100 / 10;
 						*pto++ = '0' + cnt % 10;
@@ -487,18 +435,32 @@ void substitute(struct session *ses, char *string, char *result, int flags)
 			case '@':
 				if (HAS_BIT(flags, SUB_FUN))
 				{
+					if (pti[1] == '@')
+					{
+						escape = TRUE;
+						*pto++ = *pti++;
+
+						continue;
+					}
+
 					for (ptt = temp, i = 1 ; isalnum(pti[i]) || pti[i] == '_' ; i++)
 					{
 						*ptt++ = pti[i];
 					}
 					*ptt = 0;
 
-					node = searchnode_list(ses->list[LIST_FUNCTION], temp);
+					node = search_node_list(ses->list[LIST_FUNCTION], temp);
 
 					if (node == NULL || pti[i] != DEFAULT_OPEN)
 					{
+						escape = FALSE;
 						*pto++ = *pti++;
+						continue;
+					}
 
+					if (escape)
+					{
+						pti++;
 						continue;
 					}
 
@@ -506,7 +468,7 @@ void substitute(struct session *ses, char *string, char *result, int flags)
 
 					substitute(ses, temp, buf, flags_neol);
 
-					show_debug(ses, LIST_FUNCTION, "#FUNCTION DEBUG: {%s} {%s}.", node->left, buf);
+					show_debug(ses, LIST_FUNCTION, "#DEBUG FUNCTION {%s}", node->left);
 
 					RESTRING(gtd->vars[0], buf);
 
@@ -518,20 +480,23 @@ void substitute(struct session *ses, char *string, char *result, int flags)
 
 						RESTRING(gtd->vars[i], temp);
 
-						if (*gtd->vars[i] == 0)
+						if (*pte == 0)
 						{
 							break;
 						}
+
+						if (*pte == COMMAND_SEPARATOR)
+						{
+							pte++;
+						}
 					}
 
-					pre_parse_input(ses, node->right, SUB_ARG);
+					substitute(ses, node->right, buf, SUB_ARG);
 
-					DEL_BIT(ses->flags, SES_FLAG_BREAK);
+					script_driver(ses, LIST_FUNCTION, buf);
 
 					substitute(ses, "$result", pto, flags_neol|SUB_VAR);
 
-					show_debug(ses, LIST_FUNCTION, "#FUNCTION DEBUG: {%s} {%s}.", node->left, pto);
-
 					pto += strlen(pto);
 				}
 				else
@@ -549,14 +514,7 @@ void substitute(struct session *ses, char *string, char *result, int flags)
 						{
 							*pto++ = *pti++;
 						}
-						if (isdigit(pti[1]))
-						{
-							pti++;
-						}
-						else
-						{
-							*pto++ = *pti++;
-						}
+						pti++;
 					}
 					else
 					{
@@ -581,11 +539,6 @@ void substitute(struct session *ses, char *string, char *result, int flags)
 							{
 								switch (*ptt)
 								{
-									case COMMAND_SEPARATOR:
-										*pto++ = '\\';
-										*pto++ = COMMAND_SEPARATOR;
-										break;
-
 									case '\\':
 										*pto++ = '\\';
 										*pto++ = '\\';
@@ -605,6 +558,11 @@ void substitute(struct session *ses, char *string, char *result, int flags)
 										*pto++ = 'D';
 										break;
 
+									case COMMAND_SEPARATOR:
+										*pto++ = '\\';
+										*pto++ = COMMAND_SEPARATOR;
+										break;
+
 									default:
 										*pto++ = *ptt;
 										break;
@@ -625,29 +583,6 @@ void substitute(struct session *ses, char *string, char *result, int flags)
 				}
 				break;
 
-			case '^':
-				if (HAS_BIT(flags, SUB_ANC) && (pti == string || pti[1] == 0))
-				{
-					pti++;
-				}
-				else
-				{
-					*pto++ = *pti++;
-				}
-				break;
-
-			case '~':
-				if (HAS_BIT(flags, SUB_ANC) && pti == string)
-				{
-					string++;
-					pti++;
-				}
-				else
-				{
-					*pto++ = *pti++;
-				}
-				break;
-
 			case '&':
 				if (HAS_BIT(flags, SUB_CMD) && (isdigit(pti[1]) || pti[1] == '&'))
 				{
@@ -677,6 +612,56 @@ void substitute(struct session *ses, char *string, char *result, int flags)
 						pti += isdigit(pti[2]) ? 3 : 2;
 					}
 				}
+				else if (HAS_BIT(flags, SUB_VAR) && (pti[1] == DEFAULT_OPEN || isalnum(pti[1]) || pti[1] == '&'))
+				{
+					int def = 0;
+
+					if (pti[1] == '&')
+					{
+						while (pti[1] == '&')
+						{
+							*pto++ = *pti++;
+						}
+
+						if (pti[1] == DEFAULT_OPEN || isalnum(pti[1]))
+						{
+							pti++;
+						}
+						else
+						{
+							*pto++ = *pti++;
+						}
+						continue;
+					}
+
+
+					pti++;
+
+					if (*pti == DEFAULT_OPEN)
+					{
+						def = TRUE;
+
+						pti = get_arg_in_braces(pti, buf, TRUE);
+
+						substitute(ses, buf, temp, flags_neol);
+					}
+					else
+					{
+						for (ptt = temp ; isalnum(*pti) || *pti == '_' ; i++)
+						{
+							*ptt++ = *pti++;
+						}
+						*ptt = 0;
+					}
+
+					pti = get_arg_at_brackets(pti, temp + strlen(temp));
+
+					get_nest_index(ses->list[LIST_VARIABLE], temp, buf, def);
+
+					substitute(ses, buf, pto, flags_neol - SUB_VAR);
+
+					pto += strlen(pto);
+				}
 				else
 				{
 					*pto++ = *pti++;
@@ -696,6 +681,13 @@ void substitute(struct session *ses, char *string, char *result, int flags)
 						case 'b':
 							*pto++ = '\b';
 							break;
+						case 'c':
+							if (pti[1])
+							{
+								pti++;
+								*pto++ = *pti % 32;
+							}
+							break;
 						case 'e':
 							*pto++ = '\033';
 							break;
@@ -709,16 +701,24 @@ void substitute(struct session *ses, char *string, char *result, int flags)
 							*pto++ = '\t';
 							break;
 						case 'x':
-							pti++;
-
-							if (pti[0] && pti[1])
+							if (pti[1] && pti[2])
 							{
+								pti++;
 								*pto++ = hex_number(pti);
 								pti++;
 							}
 							break;
+						case '0':
+							if (pti[1] && pti[2])
+							{
+								pti++;
+								*pto++ = oct_number(pti);
+								pti++;
+							}
+							break;
 						case '\0':
 							DEL_BIT(flags, SUB_EOL);
+							DEL_BIT(flags, SUB_LNF);
 							continue;
 						default:
 							*pto++ = *pti;
@@ -741,21 +741,28 @@ void substitute(struct session *ses, char *string, char *result, int flags)
 
 
 /******************************************************************************
-* Calls action_glob checking if the string matches, and automatically fills *
-* in the text represented by the wildcards on success. Per default the entire *
-* string is compared unless it starts or ends with the ^ character.           *
+* Calls tintin_regexp checking if the string matches, and automatically fills *
+* in the text represented by the wildcards on success.                        *
 ******************************************************************************/
 
-int check_one_action(char *line, char *original, char *action, struct session *ses)
+int check_one_regexp(struct session *ses, struct listnode *node, char *line, char *original, int option)
 {
-	char result[BUFFER_SIZE];
 	char *exp, *str;
 
-	substitute(ses, action, result, SUB_VAR|SUB_FUN|SUB_ESC);
+	if (node->regex == NULL)
+	{
+		char result[BUFFER_SIZE];
 
-	exp = result;
+		substitute(ses, node->left, result, SUB_VAR|SUB_FUN);
 
-	if (*exp == '~')
+		exp = result;
+	}
+	else
+	{
+		exp = node->left;
+	}
+
+	if (HAS_BIT(node->flags, NODE_FLAG_META))
 	{
 		exp++;
 		str = original;
@@ -763,186 +770,484 @@ int check_one_action(char *line, char *original, char *action, struct session *s
 	else
 	{
 		str = line;
-	}
+	}	
 
-	if (HAS_BIT(ses->flags, SES_FLAG_REGEXP))
-	{
-		return action_regexp(str, exp);
-	}
-
-	if (*exp == '^')
-	{
-		exp++;
-		return action_glob(ses, str, exp, 255);
-	}
-
-	if (exp[0] == '%' && isdigit(exp[1]))
-	{
-		return action_glob(ses, str, exp, 255);
-	}
-
-	while (*str)
-	{
-		if ((*exp == '[' || *exp == *str) && action_glob(ses, str, exp, 255))
-		{
-			return TRUE;
-		}
-		str++;
-	}
-	return FALSE;
+	return tintin_regexp(node->regex, str, exp, option, SUB_ARG);
 }
 
+/*
+	Keep synched with tintin_regexp and tintin_regexp_compile
+*/
 
-int action_glob(struct session *ses, char *str, char *exp, unsigned char arg)
+int tintin_regexp_check(char *exp)
 {
-	short cnt, cnt2;
-
-	thestart:
+	if (*exp == '^')
+	{
+		return TRUE;
+	}
 
 	while (*exp)
 	{
-
 #ifdef BIG5
-		if (*exp & 0x80 && *str & 0x80)
+		if (*exp & 0x80)
 		{
-			exp++;
-			str++;
-			if (*exp++ != *str++)
-			{
-				return FALSE;
-			}
+			exp += 2;
+			continue;
+		}
+#endif		
+		switch (exp[0])
+		{
+			case '\\':
+			case '{':
+				return TRUE;
+
+			case '$':
+				if (exp[1] == 0)
+				{
+					return TRUE;
+				}
+				break;
+
+			case '%':
+				switch (exp[1])
+				{
+					case '0':
+					case '1':
+					case '2':
+					case '3':
+					case '4':
+					case '5':
+					case '6':
+					case '7':
+					case '8':
+					case '9':
+					case 'd':
+					case 'D':
+					case 'i':
+					case 'I':
+					case 's':
+					case 'S':
+					case 'w':
+					case 'W':
+					case '?':
+					case '*':
+					case '+':
+					case '.':
+					case '%':
+						return TRUE;
+				}
+				break;
+		}
+		exp++;
+	}
+	return FALSE;
+}
+				
+int tintin_regexp(pcre *nodepcre, char *str, char *exp, int option, int flag)
+{
+	char out[BUFFER_SIZE], *pti, *pto;
+	int arg = 1, var = 1, fix = 0;
+
+	pti = exp;
+	pto = out;
+
+	while (*pti == '^')
+	{
+		*pto++ = *pti++;
+	}
+
+	while (*pti)
+	{
+#ifdef BIG5
+		if (*pti & 0x80)
+		{
+			*pto++ = *pti++;
+			*pto++ = *pti++;
 			continue;
 		}
 #endif
-
-		switch (exp[0])
+		switch (pti[0])
 		{
-			case '%':
-				if (isdigit(exp[1]))
-				{
-					arg = isdigit(exp[2]) ? (exp[1] - '0') * 10 + exp[2] - '0' : exp[1] - '0';
-
-					if (exp[isdigit(exp[2]) ? 3 : 2] == 0)
-					{
-						RESTRING(gtd->vars[arg], str);
-
-						return TRUE;
-					}
-
-					for (cnt = 0 ; str[cnt] ; cnt++)
-					{
-						if (action_glob(ses, &str[cnt], exp + (isdigit(exp[2]) ? 3 : 2), arg))
-						{
-							gtd->vars[arg] = refstring(gtd->vars[arg], "%.*s", cnt, str);
-
-							return TRUE;
-						}
-					}
-					return FALSE;
-				}
-				if (exp[1] == DEFAULT_OPEN)
-				{
-					arg = (arg == 99) ? arg : arg + 1;
-
-					exp = get_arg_in_braces(&exp[1], gtd->vars[arg], TRUE);
-
-					if (*exp == 0)
-					{
-						internal_variable(ses, "{%s} {%s}", gtd->vars[arg], str);
-
-						return TRUE;
-					}
-
-					for (cnt = 0 ; str[cnt] ; cnt++)
-					{
-						if (action_glob(ses, &str[cnt], exp, arg))
-						{
-							internal_variable(ses, "{%s} {%.*s}", gtd->vars[arg], cnt, str);
-
-							return TRUE;
-						}
-					}
-					return FALSE;
-				}
+			case '\\':
+				*pto++ = *pti++;
+				*pto++ = *pti++;
 				break;
 
-			case '^':
-				if (exp[1] == 0)
-				{
-					return (exp[1] == str[0]);
-				}
+			case '{':
+				gtd->args[up(var)] = up(arg);
+				*pto++ = '(';
+				pti = get_arg_in_braces(pti, pto, TRUE);
+				pto += strlen(pto);
+				*pto++ = ')';
 				break;
-				
+
 			case '[':
-				if (*str != '[' && glob(exp, "[*|*]*"))
+			case ']':
+			case '(':
+			case ')':
+			case '|':
+			case '.':
+			case '?':
+			case '+':
+			case '*':
+			case '^':
+				*pto++ = '\\';
+				*pto++ = *pti++;
+				break;
+
+			case '$':
+				if (pti[1] != DEFAULT_OPEN && !isalnum(pti[1]))
 				{
-					cnt  = 1;
-					cnt2 = 0;
+					int i = 0;
 
-					while (TRUE)
+					while (pti[++i] == '$')
 					{
-						if (exp[cnt] == '|' || exp[cnt] == ']')
-						{
-							arg = (arg == 99) ? arg : arg + 1;
-							cnt--;
-							cnt2 = 0;
+						continue;
+					}
 
-							while (exp[cnt] != '|' && exp[cnt] != '[')
-							{
-								cnt--;
-								cnt2++;
-								str++;
-							}
-							gtd->vars[arg] = refstring(gtd->vars[arg], "%.*s", cnt2, &exp[cnt+1]);
-
-							while (*exp != ']')
-							{
-								exp++;
-							}
-							exp++;
-
-							goto thestart;
-						}
-
-						if (str[cnt2] == 0)
-						{
-							break;
-						}
-
-						if (exp[cnt] == str[cnt2])
-						{
-							cnt++;
-							cnt2++;
-							continue;
-						}
-						else
-						{
-							cnt2 = 0;
-
-							while (exp[cnt])
-							{
-								if (exp[cnt] == ']')
-								{
-									return FALSE;
-								}
-								if (exp[cnt++] == '|')
-								{
-									break;
-								}
-
-							}
-						}
+					if (pti[i])
+					{
+						*pto++ = '\\';
 					}
 				}
+				*pto++ = *pti++;
+				break;
+
+			case '%':
+				switch (pti[1])
+				{
+					case '0':
+					case '1':
+					case '2':
+					case '3':
+					case '4':
+					case '5':
+					case '6':
+					case '7':
+					case '8':
+					case '9':
+						fix = SUB_FIX;
+						arg = isdigit(pti[2]) ? (pti[1] - '0') * 10 + (pti[2] - '0') : pti[1] - '0';
+						gtd->args[up(var)] = up(arg);
+						pti += isdigit(pti[2]) ? 3 : 2;
+						strcpy(pto, *pti == 0 ? "(.*)" : "(.*?)");
+						pto += strlen(pto);
+						break;
+
+					case 'd':
+						gtd->args[up(var)] = up(arg);
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "([0-9]*)" : "([0-9]*?)");
+						pto += strlen(pto);
+						break;
+
+					case 'D':
+						gtd->args[up(var)] = up(arg);
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "([^0-9]*)" : "([^0-9]*?)");
+						pto += strlen(pto);
+						break;
+
+					case 'i':
+						pti += 2;
+						strcpy(pto, "(?i)");
+						pto += strlen(pto);
+						break;
+
+					case 'I':
+						pti += 2;
+						strcpy(pto, "(?-i)");
+						pto += strlen(pto);
+						break;
+
+					case 's':
+						gtd->args[up(var)] = up(arg);
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "(\\s*)" : "(\\s*?)");
+						pto += strlen(pto);
+						break;
+
+					case 'S':
+						gtd->args[up(var)] = up(arg);
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "(\\S*)" : "(\\S*?)");
+						pto += strlen(pto);
+						break;
+
+					case 'w':
+						gtd->args[up(var)] = up(arg);
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "([a-zA-Z]*)" : "([a-zA-Z]*?)");
+						pto += strlen(pto);
+						break;
+
+					case 'W':
+						gtd->args[up(var)] = up(arg);
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "([^a-zA-Z]*)" : "([^a-zA-Z]*?)");
+						pto += strlen(pto);
+						break;
+
+					case '?':
+						gtd->args[up(var)] = up(arg);
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "(.?)" : "(.??)");
+						pto += strlen(pto);
+						break;
+
+					case '*':
+						gtd->args[up(var)] = up(arg);
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "(.*)" : "(.*?)");
+						pto += strlen(pto);
+						break;
+
+					case '+':
+						gtd->args[up(var)] = up(arg);
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "(.+)" : "(.+?)");
+						pto += strlen(pto);
+						break;
+
+					case '.':
+						gtd->args[up(var)] = up(arg);
+						pti += 2;
+						strcpy(pto, "(.)");
+						pto += strlen(pto);
+						break;
+
+					case '%':
+						*pto++ = *pti++;
+						pti++;
+						break;
+
+					default:
+						*pto++ = *pti++;
+						break;
+				}
+				break;
+
+			default:
+				*pto++ = *pti++;
 				break;
 		}
-
-		if (*exp != *str)
-		{
-			return FALSE;
-		}
-		exp++;
-		str++;
 	}
-	return TRUE;
+	*pto = 0;
+
+	return regexp_compare(nodepcre, str, out, option, flag + fix);
+}
+
+pcre *tintin_regexp_compile(struct listnode *node, char *exp, int option)
+{
+	char out[BUFFER_SIZE], *pti, *pto;
+
+	pti = exp;
+	pto = out;
+
+	if (*pti == '~')
+	{
+		pti++;
+		SET_BIT(node->flags, NODE_FLAG_META);
+	}
+
+	while (*pti == '^')
+	{
+		*pto++ = *pti++;
+	}
+
+	while (*pti)
+	{
+
+#ifdef BIG5
+		if (*pti & 0x80)
+		{
+			*pto++ = *pti++;
+			*pto++ = *pti++;
+			continue;
+		}
+#endif
+		switch (pti[0])
+		{
+			case '\\':
+				*pto++ = *pti++;
+				*pto++ = *pti++;
+				break;
+
+			case '{':
+				*pto++ = '(';
+				pti = get_arg_in_braces(pti, pto, TRUE);
+				while (*pto)
+				{
+					if (pto[0] == '$' || pto[0] == '@')
+					{
+						if (pto[1])
+						{
+							return NULL;
+						}
+					}
+					pto++;
+				}
+				*pto++ = ')';
+				break;
+
+			case '&':
+				if (pti[1] == DEFAULT_OPEN || isalnum(pti[1]) || pti[1] == '&')
+				{
+					return NULL;
+				}
+				*pto++ = *pti++;
+				break;
+
+			case '@':
+				if (pti[1] == DEFAULT_OPEN || isalnum(pti[1]) || pti[1] == '@')
+				{
+					return NULL;
+				}
+				*pto++ = *pti++;
+				break;
+
+			case '$':
+				if (pti[1] == DEFAULT_OPEN || isalnum(pti[1]))
+				{
+					return NULL;
+				}
+				{
+					int i = 0;
+
+					while (pti[++i] == '$')
+					{
+						continue;
+					}
+
+					if (pti[i])
+					{
+						*pto++ = '\\';
+					}
+				}
+				*pto++ = *pti++;
+				break;
+
+			case '[':
+			case ']':
+			case '(':
+			case ')':
+			case '|':
+			case '.':
+			case '?':
+			case '+':
+			case '*':
+			case '^':
+				*pto++ = '\\';
+				*pto++ = *pti++;
+				break;
+
+			case '%':
+				switch (pti[1])
+				{
+					case '0':
+					case '1':
+					case '2':
+					case '3':
+					case '4':
+					case '5':
+					case '6':
+					case '7':
+					case '8':
+					case '9':
+						pti += isdigit(pti[2]) ? 3 : 2;
+						strcpy(pto, *pti == 0 ? "(.*)" : "(.*?)");
+						pto += strlen(pto);
+						break;
+
+					case 'd':
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "([0-9]*)" : "([0-9]*?)");
+						pto += strlen(pto);
+						break;
+
+					case 'D':
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "([^0-9]*)" : "([^0-9]*?)");
+						pto += strlen(pto);
+						break;
+
+					case 'i':
+						pti += 2;
+						strcpy(pto, "(?i)");
+						pto += strlen(pto);
+						break;
+
+					case 'I':
+						pti += 2;
+						strcpy(pto, "(?-i)");
+						pto += strlen(pto);
+						break;
+
+					case 's':
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "(\\s*)" : "(\\s*?)");
+						pto += strlen(pto);
+						break;
+
+					case 'S':
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "(\\S*)" : "(\\S*?)");
+						pto += strlen(pto);
+						break;
+
+					case 'w':
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "([a-zA-Z]*)" : "([a-zA-Z]*?)");
+						pto += strlen(pto);
+						break;
+
+					case 'W':
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "([^a-zA-Z]*)" : "([^a-zA-Z]*?)");
+						pto += strlen(pto);
+						break;
+
+					case '?':
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "(.?)" : "(.??)");
+						pto += strlen(pto);
+						break;
+
+					case '*':
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "(.*)" : "(.*?)");
+						pto += strlen(pto);
+						break;
+
+					case '+':
+						pti += 2;
+						strcpy(pto, *pti == 0 ? "(.+)" : "(.+?)");
+						pto += strlen(pto);
+						break;
+
+					case '.':
+						pti += 2;
+						strcpy(pto, "(.)");
+						pto += strlen(pto);
+						break;
+
+					case '%':
+						*pto++ = *pti++;
+						pti++;
+						break;
+
+					default:
+						*pto++ = *pti++;
+						break;
+				}
+				break;
+
+			default:
+				*pto++ = *pti++;
+				break;
+		}
+	}
+	*pto = 0;
+
+	return regexp_compile(out, option);
 }
