@@ -45,15 +45,6 @@ int is_abbrev(char *s1, char *s2)
 	return (!strncasecmp(s2, s1, strlen(s1)));
 }
 
-int is_suffix(char *s1, char *s2)
-{
-	int sl1, sl2;
-
-	sl1 = strlen(s1);
-	sl2 = strlen(s2);
-
-	return (sl1 <= sl2 && !strcasecmp(s1, s2 + sl2 - sl1)) ? FALSE : TRUE;
-}
 
 int is_color_code(char *str)
 {
@@ -80,9 +71,8 @@ int is_color_code(char *str)
 
 int is_number(char *str)
 {
-	int i;
-
-	i = 0;
+	int i = 0;
+	int j = 0;
 
 	if (str[i] == '-' || str[i] == '+')
 	{
@@ -96,9 +86,21 @@ int is_number(char *str)
 
 	for (i = 1 ; str[i] != 0 ; i++)
 	{
-		if (!isdigit(str[i]) && str[i] != '.')
+		if (!isdigit(str[i]))
 		{
-			return FALSE;
+			if (str[i] != '.')
+			{
+				return FALSE;
+			}
+
+			for (i++ ; str[i] != 0 ; i++)
+			{
+				if (!isdigit(str[i]))
+				{
+					return FALSE;
+				}
+			}
+			return TRUE;
 		}
 	}
 	return TRUE;
@@ -106,17 +108,17 @@ int is_number(char *str)
 
 int hex_number(char *str)
 {
-	int value = '?';
+	int value = 0;
 
 	if (str)
 	{
 		if (isdigit(*str))
 		{
-			value = (*str - '0') << 4;
+			value += 16 * (*str - '0');
 		}
 		else
 		{
-			value = (*str - 'A' + 10) << 4;
+			value += 16 * (toupper(*str) - 'A' + 10);
 		}
 		str++;
 	}
@@ -129,7 +131,32 @@ int hex_number(char *str)
 		}
 		else
 		{
-			value += *str - 'A' + 10;
+			value += toupper(*str) - 'A' + 10;
+		}
+		str++;
+	}
+
+	return value;
+}
+
+int oct_number(char *str)
+{
+	int value = 0;
+
+	if (str)
+	{
+		if (isdigit(*str))
+		{
+			value += 8 * (*str - '0');
+		}
+		str++;
+	}
+
+	if (str)
+	{
+		if (isdigit(*str))
+		{
+			value += *str - '0';
 		}
 		str++;
 	}
@@ -154,22 +181,6 @@ long long utime()
 	return gtd->time;
 }
 
-char *timestamp(char *str)
-{
-	static char outbuf[BUFFER_SIZE];
-	struct tm timeval_tm;
-	time_t    timeval_t;
-
-	timeval_t = (time_t) time(NULL);
-
-	timeval_tm = *localtime(&timeval_t);
-
-	strftime(outbuf, 100, str, &timeval_tm);
-
-	return outbuf;
-}
-
-
 char *capitalize(char *str)
 {
 	static char outbuf[BUFFER_SIZE];
@@ -184,6 +195,29 @@ char *capitalize(char *str)
 	return outbuf;
 }
 
+char *ntos(long long number)
+{
+	static char outbuf[100][NUMBER_SIZE];
+	static int cnt;
+
+	cnt = (cnt + 1) % 100;
+
+	sprintf(outbuf[cnt], "%lld", number);
+
+	return outbuf[cnt];
+}
+
+char *indent(int cnt)
+{
+	static char outbuf[BUFFER_SIZE];
+
+	memset(outbuf, '\t', cnt);
+
+	outbuf[cnt] = 0;
+
+	return outbuf;
+}
+
 void cat_sprintf(char *dest, char *fmt, ...)
 {
 	char buf[STRING_SIZE];
@@ -212,6 +246,23 @@ void ins_sprintf(char *dest, char *fmt, ...)
 	strcat(dest, tmp);
 }
 
+int str_suffix(char *str1, char *str2)
+{
+	int len1, len2;
+
+	len1 = strlen(str1);
+	len2 = strlen(str2);
+
+	if (len1 >= len2)
+	{
+		if (!strcasecmp(str1 + len1 - len2, str2))
+		{
+			return FALSE;
+		}
+	}
+	return TRUE;
+}
+
 void show_message(struct session *ses, int index, char *format, ...)
 {
 	struct listroot *root;
@@ -228,7 +279,7 @@ void show_message(struct session *ses, int index, char *format, ...)
 	
 	if (index == -1)
 	{
-		if (HAS_BIT(gtd->flags, TINTIN_FLAG_SHOWMESSAGE))
+		if (HAS_BIT(gtd->flags, TINTIN_FLAG_USERCOMMAND))
 		{
 			tintin_puts2(ses, buf);
 		}
@@ -248,7 +299,7 @@ void show_message(struct session *ses, int index, char *format, ...)
 
 	if (HAS_BIT(root->flags, LIST_FLAG_MESSAGE))
 	{
-		if (HAS_BIT(gtd->flags, TINTIN_FLAG_SHOWMESSAGE))
+		if (HAS_BIT(gtd->flags, TINTIN_FLAG_USERCOMMAND))
 		{
 			tintin_puts2(ses, buf);
 
@@ -274,15 +325,16 @@ void show_debug(struct session *ses, int index, char *format, ...)
 	char buf[STRING_SIZE];
 	va_list args;
 
+	push_call("show_debug(%p,%p,%p)",ses,index,format);
+
 	root = ses->list[index];
 
 	if (!HAS_BIT(root->flags, LIST_FLAG_DEBUG) && !HAS_BIT(root->flags, LIST_FLAG_LOG))
 	{
+		pop_call();
 		return;
 	}
 
-	push_call("show_debug(%p,%p,%p)",ses,index,format);		
-
 	va_start(args, format);
 
 	vsprintf(buf, format, args);
@@ -317,7 +369,7 @@ void tintin_header(struct session *ses, char *format, ...)
 	vsprintf(arg, format, args);
 	va_end(args);
 
-	if (strlen(arg) > gtd->ses->cols - 2)
+	if ((int) strlen(arg) > gtd->ses->cols - 2)
 	{
 		arg[gtd->ses->cols - 2] = 0;
 	}
@@ -391,7 +443,7 @@ void tintin_puts3(struct session *ses, char *string)
 
 	if (strip_vt102_strlen(ses->more_output) != 0)
 	{
-		sprintf(output, "\r\n%s", string);
+		sprintf(output, "\n%s", string);
 	}
 	else
 	{
@@ -428,6 +480,8 @@ void tintin_puts2(struct session *ses, char *string)
 {
 	char output[STRING_SIZE];
 
+	push_call("tintin_puts2(%p,%p)",ses,string);
+
 	if (ses == NULL)
 	{
 		ses = gtd->ses;
@@ -435,12 +489,13 @@ void tintin_puts2(struct session *ses, char *string)
 
 	if (!HAS_BIT(gtd->ses->flags, SES_FLAG_VERBOSE) && gtd->quiet)
 	{
+		pop_call();
 		return;
 	}
 
 	if (strip_vt102_strlen(ses->more_output) != 0)
 	{
-		sprintf(output, "\r\n\033[0m%s\033[0m", string);
+		sprintf(output, "\n\033[0m%s\033[0m", string);
 	}
 	else
 	{
@@ -451,6 +506,7 @@ void tintin_puts2(struct session *ses, char *string)
 
 	if (ses != gtd->ses)
 	{
+		pop_call();
 		return;
 	}
 
@@ -466,6 +522,8 @@ void tintin_puts2(struct session *ses, char *string)
 	{
 		restore_pos(ses);
 	}
+	pop_call();
+	return;
 }
 
 /*
@@ -499,8 +557,6 @@ void tintin_puts(struct session *ses, char *string)
 
 void syserr(char *msg)
 {
-	extern int errno;
-
 	char s[128], *syserrmsg;
 
 	syserrmsg = strerror(errno);
